
import numpy as np
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
from HPO.utils.time_series_augmentation import permutation , magnitude_warp, time_warp, window_slice, jitter, scaling, rotation
import os
import matplotlib.pyplot as plt
class Repsol_Full(Dataset):

  def __init__(self, path_dir : str, max_size = 3000, min_size = 500):
    self.path = "{}/scripts/datasets/".format(os.environ["HOME"])+path_dir
    data = []
    DEBUG = False
    self.x_index_address = {}
    self.y_index_address = {}
    #index_list sum()
    files = os.listdir(self.path)
    self.device_track = {}
    for c,i in enumerate(files):
      sample = np.reshape(np.load(self.path+i),(-1,28))
      if DEBUG:
        plt.plot(sample)
        plt.savefig("datadebug") 
        input("waiting for next plot")
      if sample.shape[0] > min_size:
        self.device_track[c] = i[:2]
        data.append(sample[-1*max_size:,:])
    self.current_index = 0
    self.last_batch = []
    self.real_samples = len(data)
    print(self.real_samples)
    for datapoint, batch in enumerate(data):
      batch_x = batch[:,1:]
      batch_y = batch[:,0]
      self.add_to_dataset(batch_x,batch_y)
      self.real_data_index = self.current_index
    self.n_samples = self.current_index #* len(self.augmentations)
    self.features = 27
    self.n_classes = 2

    
  def save_2_file(self, x,y, datapoint , aug_num):
    arr = np.zeros((x.shape[0],x.shape[1]+ 1))
    arr[:,0] = y
    arr[:,1:] = x
    np.save(self.path+"repsol_augmented/{}-{}".format(datapoint, aug_num), arr)
  def get_linked(self, idx):
    return self.aug_source_dict[idx]

  def add_to_dataset(self,x,y):

    self.x_index_address[self.current_index] = torch.from_numpy(np.swapaxes(x, 0,1))
    self.y_index_address[self.current_index] = torch.from_numpy(np.unique(y).reshape((1,)))
    self.current_index += 1 

  def set_window_size(self, window_size):
    self.window = window_size

  def __getitem__(self, index):
    #index_address keys are the first usable index in a batch for the window size
    #this means the index
    x = self.x_index_address[index]
    y = self.y_index_address[index]
    self.last_batch.append(index)   

    return x , y
  
  def y(self):#
    out = []
    for i in self.y_index_address:
      out.append(self.y_index_address[i].numpy())
    return out
  def get_source(self):
    sources = []
    for i in self.last_batch:
      sources.append(self.device_track[i])
    self.last_batch = []
    return sources
  
  def get_id(self):
    return self.last_batch
    

  def get_n_classes(self):
    return self.n_classes
  
  def get_n_features(self):
    return self.features
  def get_n_samples(self):
    return self.n_samples

  def __len__(self):
    return self.n_samples 



class Mixed_repsol_full(Repsol_Full):
  def __init__(self, path_dir = None):
    path = "{}/scripts/datasets/repsol_mixed".format(os.environ["HOME"])
    if path_dir == None: 
      path_dir = "repsol_mixed/"
    super().__init__( path_dir)


class repsol_unlabeled(Dataset):
  def __init__(self, window_size = 500):
    self.soft_labels = False
    path = "{}/scripts/datasets/repsol_unlabeled/".format(os.environ["HOME"])
    data = []
    self.index_address = {}

    files = os.listdir(path)

    for i in files:
      data.append(np.reshape(np.load(path+i),(-1,27)))
    self.current_index = 0


  
    for batch in data:
      count = window_size
      if batch.shape[0] != 0:
        while batch.shape[0] - count > window_size:
          self.add_to_dataset(batch[count-window_size:count,:])
          count += window_size

      
    
    self.n_samples = self.current_index #* len(self.augmentations)
    self.features = 27
    self.n_classes = 2
  def add_to_dataset(self,x):
    
    self.index_address[self.current_index] = torch.from_numpy(x.reshape(x.shape[1], x.shape[0]))
    self.current_index += 1 

  def add_labels(self , y):
    self.y_index_address = {}
    current_index = 0
    for y_s in y:
      self.y_index_address[current_index] = torch.from_numpy(y_s)
      current_index += 1 
    if current_index != self.current_index:
      raise ValueError("Incorrect number of labels")
    self.soft_labels = True


  def set_window_size(self, window_size):
    self.window = window_size

  def __getitem__(self, index):
    #index_address keys are the first usable index in a batch for the window size
    #this means the index
    if self.soft_labels == False:
      x = self.index_address[index]
      return x 
    else:
      x = self.index_address[index]
      x = self.y_index_address[index]
      return x, y
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples

  def __len__(self):
    return self.n_samples 



class repsol_feature(Dataset):
  def __init__(self,name):
    self.path = "{}/scripts/datasets/repsol_features/".format(os.environ["HOME"])
    self.data = np.load(self.path+name)
      
    self.n_classes = 2
    self.n_samples = self.data.shape[0]
    self.n_features = self.data.shape[1] - 1

  def __getitem__(self, index):
    #index_address keys are the first usable index in a batch for the window size
    #this means the index
    x = self.data[index,1:]
    y = self.data[index,0]
    return x , y
  
  def get_n_classes(self):
    return self.n_classes
  
  def get_n_features(self):
    return self.n_features

  def get_n_samples(self):
    return self.n_samples
  
  def __len__(self):
    return self.n_samples 
    


class Train_repsol_feature(repsol_feature):
  def __init__(self): 
    super().__init__("train_selected.npy")

class Test_repsol_feature(repsol_feature):
  def __init__(self): 
    super().__init__("test_selected.npy")

class Mixed_repsol_feature(repsol_feature):
  def __init__(self): 
    super().__init__("full_selected.npy")




import os
import random
import time
import shutil
from sklearn.model_selection import StratifiedKFold as KFold
class MetaCV:
  def __init__(self, source_path, destination_path, num_samples, num_folds = 3, log_file = "log.txt"):
    assert(num_samples == len(os.listdir(source_path))), "Number given samples {} does not match number of files in directory {}".format(num_samples,len(os.listdir(source_path)))
    fold_size = int(num_samples/num_folds)
    self.source_path = source_path
    self.destination_path = destination_path
    self.log_file = log_file
    self.current_fold = 0
    self.sample_files = os.listdir(source_path)
    self.num_folds = num_folds
    self.seed = random.randint(0,9999)
    #This is where the active samples will be stored.
    self.main_train_path = "{}/train".format(self.destination_path)
    self.main_test_path = "{}/test".format(self.destination_path)
    if not os.path.exists(destination_path):
      os.mkdir(destination_path)
    if not os.path.exists(self.main_train_path):
      os.mkdir(self.main_train_path) 
    if not os.path.exists(self.main_test_path):
      os.mkdir(self.main_test_path) 
    self.get_classes() 
    self.kfold = KFold(n_splits = self.num_folds, shuffle = True,random_state = self.seed)
    self.fold = self.kfold.split(self.sample_files,self.class_list)

  def get_classes(self):
    self.class_list = []
    for i in self.sample_files:
      self.class_list.append(i[3:11])
      
    print("List of classes: {}".format(set(self.class_list)))
 
  def next_fold(self):
    #Set up storing of samples
    assert(self.current_fold < self.num_folds), "[Error]: Tried to access fold {} when there are only {} folds!".format(self.current_fold, self.num_folds)
    self.check_for_overwrite(self.destination_path,self.current_fold)
    print("Starting Meta Cross Validation fold: {}".format(self.current_fold))
    train_path = "{}/cv-{}/train".format(self.destination_path,self.current_fold)
    test_path = "{}/cv-{}/test".format(self.destination_path,self.current_fold)
    self.clear_main_dir()
    os.mkdir("{}/cv-{}".format(self.destination_path,self.current_fold)) 
    os.mkdir(train_path) 
    os.mkdir(test_path) 
    current_fold_train_indexes, current_fold_test_indexes = next(self.fold)
    
    for file_index in current_fold_train_indexes:
      self.move_file(train_path,file_index)
      self.move_file(self.main_train_path,file_index)

    for file_index in current_fold_test_indexes:
      self.move_file(test_path,file_index)
      self.move_file(self.main_test_path,file_index)
    self.current_fold += 1

  def clear_main_dir(self):
      train_files = os.listdir(self.main_train_path)
      test_files = os.listdir(self.main_test_path)
      for i in train_files:
        os.remove("{}/{}".format(self.main_train_path,i)) 
      for i in test_files:
        os.remove("{}/{}".format(self.main_test_path,i)) 
       
  def move_file(self,path, idx):
      shutil.copyfile("{}/{}".format(self.source_path,self.sample_files[idx]) , "{}/{}".format(path,self.sample_files[idx]))

  def check_for_overwrite(self,dest, fold):
    while fold < self.num_folds:
      if os.path.exists("{}/cv-{}".format(dest, fold)):
        ans = input("folder for cv-{} already exists overwrite? [Y/n]".format(fold))
        if ans.upper() == "Y":
          if os.path.exists("{}/cv-{}-old".format(dest, fold)):
            os.system("rm -r {}/cv-{}-old".format(dest, fold))
          os.rename("{}/cv-{}".format(dest, fold),"{}/cv-{}-old".format(dest, fold))
          with open("{}/{}".format(self.destination_path, self.log_file), "w") as text_file:
            text_file.write(str(self.seed))
        else:
          print("Loading previous run")
          with open("{}/{}".format(self.destination_path, self.log_file), "r") as text_file:
            self.seed = int(text_file.read())
            self.kfold = KFold(n_splits = self.num_folds, shuffle = True,random_state = self.seed)
            self.fold = self.kfold.split(self.sample_files,self.class_list)
            folds = os.listdir(self.destination_path)
            max_val = 0
            for i in folds:
              if i[-1].isnumeric:
                if int(i[-1]) > max_val:
                  max_val = int(i[-1])
            self.current_fold = max_val

            
      else:
        with open("{}/{}".format(self.destination_path, self.log_file), "w") as text_file:
          text_file.write(str(self.seed))
      fold += 1




PRIMITIVES = {
    # macro level
    "Linear1": [(1,2)],
    "Linear3": [(1,2),(2,3),(3,4)],
    "Linear4": [(1,2),(2,3),(3,4),(4,5),
    "Residual3": [(1, 2), (2, 3), (1, 4), (3, 4)],
    "Diamond3": [(1, 2), (1, 3), (2, 4), (3, 5), (4, 6), (5, 6)],
    "Linear2": [(1,2),(2,3)],
    "Residual2": [(1,2),(2,3),(1,3)],
    "Diamond2": [(1,2),(1,3),(3,4),(2,4)],
    "Cell" : [(1, 2), (1, 3), (2, 3), (1, 4), (2, 4), (3, 4)]
    # cell level
    "id": ops.Identity(),
    "zero": ops.Zero(stride=1),
    "conv3x3": {
        "op": nb201_ops.ReLUConvBN,
        "kernel_size": 3,
        "stride": 1,
        "padding": 1,
        "dilation": 1,
        "affine": True,
    },
    "conv1x1": {
        "op": nb201_ops.ReLUConvBN,
        "kernel_size": 1,
        "stride": 1,
        "padding": 0,
        "dilation": 1,
        "affine": True,
    },
    "avg_pool": {"op": nb201_ops.POOLING, "mode": "avg", "stride": 1, "affine": True},
    "Cell": NASBench201Cell,
    # conv block level
    "conv3x3o": {
        "op": nb201_ops.Conv,
        "kernel_size": 3,
    },
    "conv1x1o": {"op": nb201_ops.Conv, "kernel_size": 1},
    "dconv3x3o": {
        "op": nb201_ops.DepthwiseConv,
        "kernel_size": 3,
    },
    "batch": {"op": nb201_ops.Normalization, "norm_type": "batch_norm"},
    "instance": {
        "op": nb201_ops.Normalization,
        "norm_type": "instance_norm",
    },
    "layer": {"op": nb201_ops.Normalization, "norm_type": "layer_norm"},
    "relu": {"op": nb201_ops.Activation, "act_type": "relu"},
    "hardswish": {"op": nb201_ops.Activation, "act_type": "hardswish"},
    "mish": {"op": nb201_ops.Activation, "act_type": "mish"},
}


import torch
import torch.nn as nn

OPS = {
  'none' : lambda C, stride, affine: Zero(stride),
  'avg_pool_3x3' : lambda C, stride, affine: nn.AvgPool2d(3, stride=stride, padding=1, count_include_pad=False),
  'max_pool_3x3' : lambda C, stride, affine: nn.MaxPool2d(3, stride=stride, padding=1),
  'skip_connect' : lambda C, stride, affine: Identity() if stride == 1 else FactorizedReduce(C, C, affine=affine),
  'sep_conv_3x3' : lambda C, stride, affine: SepConv(C, C, 3, stride, 1, affine=affine),
  'sep_conv_5x5' : lambda C, stride, affine: SepConv(C, C, 5, stride, 2, affine=affine),
  'sep_conv_7x7' : lambda C, stride, affine: SepConv(C, C, 7, stride, 3, affine=affine),
  'dil_conv_3x3' : lambda C, stride, affine: DilConv(C, C, 3, stride, 2, 2, affine=affine),
  'dil_conv_5x5' : lambda C, stride, affine: DilConv(C, C, 5, stride, 4, 2, affine=affine),
  'conv_7x1_1x7' : lambda C, stride, affine: nn.Sequential(
    nn.ReLU(inplace=False),
    nn.Conv2d(C, C, (1,7), stride=(1, stride), padding=(0, 3), bias=False),
    nn.Conv2d(C, C, (7,1), stride=(stride, 1), padding=(3, 0), bias=False),
    nn.BatchNorm2d(C, affine=affine)
    ),
}

class ReLUConvBN(nn.Module):

  def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
    super(ReLUConvBN, self).__init__()
    self.op = nn.Sequential(
      nn.ReLU(inplace=False),
      nn.Conv2d(C_in, C_out, kernel_size, stride=stride, padding=padding, bias=False),
      nn.BatchNorm2d(C_out, affine=affine)
    )

  def forward(self, x):
    return self.op(x)

class DilConv(nn.Module):
    
  def __init__(self, C_in, C_out, kernel_size, stride, padding, dilation, affine=True):
    super(DilConv, self).__init__()
    self.op = nn.Sequential(
      nn.ReLU(inplace=False),
      nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, groups=C_in, bias=False),
      nn.Conv2d(C_in, C_out, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm2d(C_out, affine=affine),
      )

  def forward(self, x):
    return self.op(x)


class SepConv(nn.Module):
    
  def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
    super(SepConv, self).__init__()
    self.op = nn.Sequential(
      nn.ReLU(inplace=False),
      nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, groups=C_in, bias=False),
      nn.Conv2d(C_in, C_in, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm2d(C_in, affine=affine),
      nn.ReLU(inplace=False),
      nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=1, padding=padding, groups=C_in, bias=False),
      nn.Conv2d(C_in, C_out, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm2d(C_out, affine=affine),
      )

  def forward(self, x):
    return self.op(x)


class Identity(nn.Module):

  def __init__(self):
    super(Identity, self).__init__()

  def forward(self, x):
    return x


class Zero(nn.Module):

  def __init__(self, stride):
    super(Zero, self).__init__()
    self.stride = stride

  def forward(self, x):
    if self.stride == 1:
      return x.mul(0.)
    return x[:,:,::self.stride,::self.stride].mul(0.)


class FactorizedReduce(nn.Module):

  def __init__(self, C_in, C_out, affine=True):
    super(FactorizedReduce, self).__init__()
    assert C_out % 2 == 0
    self.relu = nn.ReLU(inplace=False)
    self.conv_1 = nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
    self.conv_2 = nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False) 
    self.bn = nn.BatchNorm2d(C_out, affine=affine)

  def forward(self, x):
    x = self.relu(x)
    out = torch.cat([self.conv_1(x), self.conv_2(x[:,:,1:,1:])], dim=1)
    out = self.bn(out)
    return out




import numpy as np
import torch.nn.functional as F
import os
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt

class EEG(Dataset):

  def __init__(self, train: bool, augmentation = None, cuda_device = 0,**kwargs):
    PATH = "/home/cmackinnon/scripts/datasets/EEG/"
    self.cuda_device = cuda_device
    if train:
      PATH+="train/"
    else:
      PATH+="test/"

    self.x = torch.from_numpy(np.load(PATH+"x.npy")).cuda(cuda_device).float()
    self.y = torch.from_numpy(np.load(PATH+"y.npy"))
    self.groups = np.load(PATH+"groups.npy")
    if kwargs["binary"]:
      self.y = torch.from_numpy(np.where(self.y.numpy() != 0, 1,0)).cuda(cuda_device).long()
    else:
      self.y = self.y.cuda(cuda_device).long()
    self.augmentation = augmentation
  def get_labels(self):
    return self.y
  def disable_augmentation(self):
    self.augmentation = False
  def enable_augmentation(self,aug):
    self.augmentation = aug

  def __getitem__(self, index):
    x, y = self.x[index], self.y[index]
    if self.augmentation:
      for func in self.augmentation:
        x,y = func(x,y)
    return x,y
  

  def get_groups(self, train_id, test_id):
      train_groups = self.groups[train_id]
      test_groups = self.groups[test_id]
      print("Groups in train are: {} with a total length: {}".format(np.unique(train_groups),len(train_groups)))
      print("Groups in test are: {} with a total length: {}".format(np.unique(test_groups),len(test_groups)))

  def get_n_classes(self):
    return len(torch.unique(self.y))
  def get_n_samples(self):
    return self.x.shape[0]
  def get_n_features(self):
    return self.x.shape[1]
  def get_length(self):
    return self.x.shape[2]

  def __len__(self):
    return self.x.shape[0]


class Train_EEG(EEG):

  def __init__(self, window_size = 500, augmentation = False,samples_per_class = None,binary = False,one_hot = True,samples_per_epoch = 1,PATH= None,cuda_device = None,sub_set_classes = None,**kwargs): 
    super(Train_EEG,self).__init__(True, augmentation,samples_per_class = samples_per_class,binary = binary ,one_hot = one_hot,samples_per_epoch = samples_per_epoch,PATH = PATH,cuda_device = cuda_device,sub_set_classes = sub_set_classes,**kwargs)

class Test_EEG(EEG):

  def __init__(self, window_size = 500, augmentation = False,samples_per_class = None,binary = False,one_hot = False,samples_per_epoch = 1,cuda_device = None,sub_set_classes = None,**kwargs): 
    super(Test_EEG,self).__init__(False , augmentation,samples_per_class =samples_per_class,binary = binary ,one_hot = one_hot,samples_per_epoch = samples_per_epoch,cuda_device = cuda_device,sub_set_classes = sub_set_classes,**kwargs)




import random
import torch
import torch.nn as nn
import numpy as np
from torch import Tensor
from HPO.utils.model_constructor import Model
from torch.utils.data import DataLoader
from HPO.utils.time_series_augmentation_torch import jitter, scaling , rotation, window_warp, crop, cutout,mix_up, cut_mix
class WeightTest:
  def __init__(self, model):
    self.count_dict = {}
    self.model = model
    for name, param in self.model.named_parameters():
      if 'weight' in name:
        self.count_dict[name] = 0
  def step(self):
    for name, param in self.model.named_parameters():
      if ('weight' in name) and (name != "nets.0.fc.weight"):
        print("Name: {}".format(name))
        temp = torch.zeros(param.grad.shape)
        temp[param.grad != 0] += 1
        self.count_dict[name] += temp
        print("Weight {} updates: {}".format(name, min(self.count_dict[name])))


import gc
def tensor_list():
  for obj in gc.get_objects():
      try:
          if torch.is_tensor(obj) or (hasattr(obj, 'data') and torch.is_tensor(obj.data)):
              print(type(obj), obj.size())
      except:
          pass

class BarlowTwins(nn.Module):
  def __init__(self, model):
     super().__init__()
     layers = []
     self.nets = nn.ModuleList()
     self.nets.append(model)
     self.model = model
     size = model.get_channels()
     for i in range(2):
         if i ==0:
           layers.append(nn.Linear(size, size*4, bias=False))
         else:
           layers.append(nn.Linear(size*4, size*4, bias=False))
         layers.append(nn.BatchNorm1d(size*4))
         layers.append(nn.ReLU(inplace=True))
     layers.append(nn.Linear(size*4, size*4, bias=False))
     self.projector = nn.Sequential(*layers)
     self.nets.append(self.projector)
  def forward(self , x): 
    x = self.nets[0]._forward(x)
    x = self.nets[1](x)
    return x

class Metaloader:
  def __init__(self, dataloaders,max_epochs):
    pass  

  def __getitem__(self,idx):
    non_empty_dataloaders = list(dataloaders)
    iter_dl = {}
    for b in dataloaders:
      iter_dl[b] = iter(dataloaders[b])
    for i in range(n_iter):
      while True:
        batch_size = random.choice(non_empty_dataloaders)
        try:
          samples, labels = next(iter_dl[batch_size])
        except StopIteration:
          non_empty_dataloaders.remove(batch_size)
        else:
          break



def stdio_print_training_data( iteration : int , outputs : Tensor, labels : Tensor , epoch : int, epochs : int, correct :int , total : int , peak_acc : float , loss : Tensor, n_iter, loss_list = None, binary = True):
  def cal_acc(y,t):
    c = np.count_nonzero(y.T==t)
    tot = len(t)
    return c , tot

  def convert_label_max_only(y):
    y = y.cpu().detach().numpy()
    idx = np.argmax(y,axis = 1)
    out = np.zeros_like(y)
    for count, i in enumerate(idx):
        out[count, i] = 1
    return idx
  if binary == True:
    new_correct , new_total = cal_acc(outputs.ge(0.5).cpu().detach().numpy(), labels.cpu().detach().numpy())
  else:
    new_correct , new_total =  cal_acc(convert_label_max_only(outputs), labels.cpu().detach().numpy())
  correct += new_correct 
  total += new_total
  acc = correct / total 
  if acc > peak_acc:
    peak_acc = acc
  # Save the canvas
  print("Epoch (",str(epoch),"/",str(epochs), ") Accuracy: ","%.2f" % acc, "Iteration(s) (", str(iteration),"/",str(n_iter), ") Loss: "
    ,"%.2f" % loss," Correct / Total : {} / {} ".format(correct , total),  end = '\r')
  return correct ,total ,peak_acc

def train_model_multibatch(model : Model , hyperparameter : dict, dataloaders : dict , epochs : int, batch_size : int, cuda_device = None, augment_num = 2, graph = None, binary = False):
  """
  Trains models with an assortment of batch sizes to produce a wider range of augmentated samples
  """

  if cuda_device == None:
    cuda_device = torch.cuda.current_device()
  optimizer = torch.optim.Adam(model.parameters(),lr = hyperparameter["lr"])
  #scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size = hyperparameter["lr_step"])
  #scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer,T_0 = hyperparameter["T_0"] , T_mult = hyperparameter["T_mult"])
  if binary == True:
    criterion = nn.BCEWithLogitsLoss().cuda(device = cuda_device)
  
  else:
    criterion = nn.CrossEntropyLoss()
  epoch = 0
  peak_acc = 0
  loss_list = []
  total = 0
  correct = 0
  while epoch < epochs:
    if epoch % 3 == 0:
      total = 0
      correct = 0
    aug_list = []
    aug_labels = []
    non_empty_dataloaders = list(dataloaders)
    iter_dl = {}
    for b in dataloaders:
      iter_dl[b] = iter(dataloaders[b])
    if epoch < epochs/2:
      non_empty_dataloaders.remove(4)
    if epoch < epochs/2:
      non_empty_dataloaders.remove(8)
    if epoch < 2*(epochs/3):
      non_empty_dataloaders.remove(16)
    if epoch > 3*(epochs/4):
      non_empty_dataloaders.remove(2)
    n_iter = 0
    for b_s in non_empty_dataloaders:
      n_iter += len(dataloaders[b_s])
    for i in range(n_iter):
      while True:
        batch_size = random.choice(non_empty_dataloaders)
        try:
          samples, labels = next(iter_dl[batch_size])
        except StopIteration:
          non_empty_dataloaders.remove(batch_size)
        else:
          break
      samples = samples.cuda(non_blocking=True, device = cuda_device)
      labels = labels.cuda(non_blocking=True, device = cuda_device)
      aug_sample , aug_label = augment(samples,labels,hyperparameter, cuda_device)
      aug_list.append(aug_sample)
      aug_labels.append(aug_label)
      optimizer.zero_grad()
      if batch_size > 1:
        labels = labels.long().view( batch_size  )
      else:
        labels = labels.long().view( 1 )
      outputs = model(samples.float()).cuda(device = cuda_device)
      # forward + backward + optimize
      if binary:
        loss = criterion(outputs.view(batch_size), labels.float()).cuda(device = cuda_device)
      else:
        loss = criterion(outputs, labels).cuda(device = cuda_device)
      loss.backward()
      loss_list.append(loss.item())
      optimizer.step()
      if i %5 == 0:
        if graph != None:
          graph.put(loss_list)
        correct , total, peak_acc = stdio_print_training_data(i , outputs , labels, epoch,epochs , correct , total, peak_acc, loss.item(), n_iter, loss_list, binary = binary)


    #tensor_list()
    #scheduler.step()
    epoch += 1 
    #dataloader.set_iterator()
  print()
  print("Num epochs: {}".format(epoch))

def train_model_aug(model : Model , hyperparameter : dict, dataloader : DataLoader , epochs : int, batch_size : int, cuda_device = None, augment_num = 2, graph = None, binary = False):
  if cuda_device == None:
    cuda_device = torch.cuda.current_device()
  n_iter = len(dataloader) 
  optimizer = torch.optim.Adam(model.parameters(),lr = hyperparameter["lr"])
  #scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size = hyperparameter["lr_step"])
  #scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer,T_0 = hyperparameter["T_0"] , T_mult = hyperparameter["T_mult"])
  if binary == True:
    criterion = nn.BCEWithLogitsLoss().cuda(device = cuda_device)
  
  else:
    criterion = nn.CrossEntropyLoss()
  epoch = 0
  peak_acc = 0
  loss_list = []
  total = 0
  correct = 0
  while epoch < epochs:
    if epoch % 3 == 0:
      total = 0
      correct = 0
    aug_list = []
    aug_labels = []
    for i, (samples, labels) in enumerate( dataloader ):
      samples = samples.cuda(non_blocking=True, device = cuda_device)
      labels = labels.cuda(non_blocking=True, device = cuda_device)
      if augment_num:
          aug_sample , aug_label = augment(samples,labels,hyperparameter, cuda_device)
          aug_list.append(aug_sample)
          aug_labels.append(aug_label)
      else:
          aug_list.append(samples)
          aug_labels.append(labels)

    for i , (samples , labels) in enumerate(zip(aug_list , aug_labels)):
      optimizer.zero_grad()
      if batch_size > 1:
        labels = labels.long().view( batch_size  )
      else:
        labels = labels.long().view( 1 )
      outputs = model(samples.float()).cuda(device = cuda_device)
      # forward + backward + optimize
      if binary:
        loss = criterion(outputs.view(batch_size), labels.float()).cuda(device = cuda_device)
      else:
        loss = criterion(outputs, labels).cuda(device = cuda_device)
      loss.backward()
      loss_list.append(loss.item())
      optimizer.step()
      if i %5 == 0:
        if graph != None:
          graph.put(loss_list)
        correct , total, peak_acc = stdio_print_training_data(i , outputs , labels, epoch,epochs , correct , total, peak_acc, loss.item(), n_iter, loss_list,binary = binary)


    #tensor_list()
    #scheduler.step()
    epoch += 1 
    #dataloader.set_iterator()
  print()
  print("Num epochs: {}".format(epoch))

def train_model(model : Model , hyperparameter : dict, dataloader : DataLoader , epochs : int, 
    batch_size : int, cuda_device = None, augment_on = 0, graph = None, binary = False,evaluator= None):
  if cuda_device == None:
    cuda_device = torch.cuda.current_device()
  n_iter = len(dataloader) 
  optimizer = torch.optim.Adam(model.parameters(),lr = hyperparameter["lr"])
  #scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size = hyperparameter["lr_step"])
  #scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer,T_0 = hyperparameter["T_0"] , T_mult = hyperparameter["T_mult"])
  if binary == True:
    criterion = nn.BCEWithLogitsLoss().cuda(device = cuda_device)
  
  else:
    #if "c1" in hyperparameter:
    #  criterion = nn.CrossEntropyLoss(torch.Tensor([1, hyperparameter["c1"]])).cuda(device = cuda_device)
    #else:
    criterion = nn.CrossEntropyLoss()
  epoch = 0
  peak_acc = 0
  loss_list = []
  total = 0
  correct = 0
  while epoch < epochs:
    if epoch % 3 == 0:
      total = 0
      correct = 0
    for i, (samples, labels) in enumerate( dataloader ):
      #samples = samples.cuda(non_blocking=True, device = cuda_device)
      #labels = labels.cuda(non_blocking=True, device = cuda_device)
      batch_size = samples.shape[0]
      # zero the parameter gradients
      optimizer.zero_grad()
      #if batch_size > 1:
      #  labels = labels.long().view( batch_size  )
      #else:
      #  labels = labels.long().view( 1 )
      outputs = model(samples.float()).cuda(device = cuda_device)
      # forward + backward + optimize
      if batch_size == 1:
        outputs = outputs.view(batch_size,outputs.shape[0])
      if binary:
        loss = criterion(outputs.view(batch_size), labels.float()).cuda(device = cuda_device)
      else:
        loss = criterion(outputs, labels).cuda(device = cuda_device)
      loss.backward()
      loss_list.append(loss.item())
      optimizer.step()
      if i %10 == 0:
        if graph != None:
          graph.put(loss_list)

      if i % 500 == 0 and i != 0:
        if evaluator != None:
          evaluator.forward_pass(model,subset = 200)
          evaluator.predictions(model_is_binary = False)
          acc  =  evaluator.T_ACC()
          evaluator.reset_cm()
          print("Validation set Accuracy: {}".format(acc))
      if i% 5 == 0:
        #correct , total, peak_acc = stdio_print_training_data(i , outputs , labels, epoch,epochs , correct , total, peak_acc, loss.item(), n_iter, loss_list)
        correct , total, peak_acc = stdio_print_training_data(i , outputs , labels, epoch,epochs , correct , total, peak_acc, loss.item(), n_iter, loss_list,binary = binary)


    #scheduler.step()
    epoch += 1
    #dataloader.set_iterator()
  print()
  print("Num epochs: {}".format(epoch))



def train_model_bt(model : Model , hyperparameter : dict, dataloader : DataLoader , epochs : int, batch_size : int, cuda_device = None, validation_set = None, graph = None):
  if cuda_device == None:
    cuda_device = torch.cuda.current_device()
  n_iter = len(dataloader) 
  bt = BarlowTwins(model).cuda(device = cuda_device)
  optimizer = torch.optim.Adam(model.parameters(),lr = hyperparameter["lr"])
  criterion = barlow_twins
  epoch = 0
  loss_list = []
  #weight_test = WeightTest(bt)
  peak_acc = 0
  while epoch < epochs:
    for i, samples in enumerate( dataloader ):
      # zero the parameter gradients
      optimizer.zero_grad()
      loss = criterion(bt , samples, cuda_device = cuda_device)
      # forward + backward + optimize
      loss.backward()
      loss_list.append(loss.item())
      #weight_test.step()
      optimizer.step()
      if i %3 == 0:
        if graph != None:
          graph.put(loss_list)
        print("Epoch: {}/{} - Iteration: {}/{} - Loss: {} ".format(epoch , epochs,i, n_iter , loss.item()))
    epoch += 1 
    #dataloader.set_iterator()
  print()
  print("Num epochs: {}".format(epoch))





def augment(x, y,hp, device = None):
  augs = [jitter, scaling, window_warp,crop, mix_up,cutout]

  if "jitter" in hp:
    args = [hp["jitter"], hp["scaling"], hp["window_warp_num"], hp["crop"] , hp["mix_up"], hp["cut_out"]]
    rates = [hp["jitter_rate"], hp["scaling_rate"], hp["window_warp_rate"],hp["crop_rate"], hp["mix_up_rate"], hp["cut_out_rate"]]

    if device == None:
      for func,arg,rate in zip(augs,args, rates):

        if rate > 1:
          for _ in range((int(rate)+1)):
            if random.random() < (rate/(int(rate)+1)):
              x,y = func(x, y, arg )
        else:
          if random.random() < rate:
            x,y = func(x,y, arg )

    else:
      for func,arg,rate in zip(augs,args, rates):
        if rate > 1:
          for _ in range((int(rate)+1)):
            if random.random() < (rate/(int(rate)+1)):
              x,y = func(x, y, arg )
        else:
          if random.random() < rate:
            x,y = func(x,y, arg , device = device)

    return x,y

  else:
    rate = 0.5
    if device == None:
      for func in augs:
        if random.random() < rate:
          x,y = func(x,y)
    else:
      for func in augs:
        if random.random() < rate:
          x,y = func(x,y, device = device)
    return x ,y 
def barlow_twins(model, batch, cuda_device = None):
  N = batch.shape[0] #Batch Size
  lmbda = 0.005
  #Generate Augmented
  y_a = augment(batch)  
  y_b = augment(batch)
  
  #Generate network embeddings
  z_a = model(y_a.cuda(device = cuda_device).float())
  z_b = model(y_b.cuda(device = cuda_device).float())
  D = z_a.shape[1] #Output Dim

  #Normalise
  z_a_norm = (z_a - z_a.mean(axis = 0)) / z_a.std(axis = 0)
  z_b_norm = (z_b - z_b.mean(axis = 0)) / z_b.std(axis = 0)

  c = torch.matmul(z_a_norm.T , z_b_norm) / N
  idn = torch.eye(D).cuda(device = cuda_device)
  c_diff = torch.pow((c - idn), 2)
  
  #Seperate diag
  c_diff_diag = torch.matmul(c_diff, idn)
  c_diff_non_diag = torch.matmul( c_diff , torch.ones(D,D).cuda(device = cuda_device) - idn )
  c_diff_non_diag = c_diff_non_diag * lmbda  
  c_diff = c_diff_diag + c_diff_non_diag

  loss = c_diff.sum()
  
  return loss









# -*- coding: utf-8 -*-
"""
Pytorch triplet network functions

Created on Thu Jun 16 12:20:41 2022

@author: jackw

Adapted from:
    https://github.com/omoindrot/tensorflow-triplet-loss/blob/master/model/triplet_loss.py
    
"""

import torch as T
import time


# ---------------------------- Helper Functions -------------------------------

# calculate euclidean distance matrix for online mining 
def distance_matrix(embeddings, squared = True):
    
    #find the dot product matrix for embeddings
    dot_product = T.matmul(embeddings,T.transpose(embeddings, 0, 1))

    #get digonals from dp matrix (squared sum of embedding row)
    sq_sum = T.diagonal(dot_product, 0)

    
    '''
    calculate the distance matrix using: 
        ||a - b||^2 = ||a||^2  - 2 <a, b> + ||b||^2
    
    shape is shape (batch_size, batch_size)
    '''
    
    d =  T.unsqueeze(sq_sum, 0) - 2.0 * dot_product + T.unsqueeze(sq_sum, 1)
    d = T.clamp(d, min = 0.0)  #prevent negatives due to floating point errors
    
    
    #sqrt distance matrix if squared distance not required
    if not squared:
        mask = T.eq(d, T.tensor(0.0)).float()  #find zero value s
        d = d + (mask * 1e-16)  #replace with small epsilon value

        d = T.sqrt(d)  #sqrt distance matrix
        
        d = d * (1.0 - mask)  #set epsilon values back to 0.0

    return d


#get masks for triplet mining
def _get_triplet_mask(labels, device = 'cpu'):
    """Return a 3D mask where mask[a, p, n] is True iff the triplet (a, p, n) is valid.
    A triplet (i, j, k) is valid if:
        - i, j, k are distinct
        - labels[i] == labels[j] and labels[i] != labels[k]
    Args:
        labels: tf.int32 `Tensor` with shape [batch_size]
    """
    
    # Check that i, j and k are distinct
    I_eq = T.eye(labels.shape[0], dtype = T.bool).to(device)  #get matrix where indices are equal (BxB boolean identity matrix)
    I_neq = T.logical_not(I_eq)  #invert matrix to find where indices are not equal 
    
    z_neq_r = T.unsqueeze(I_neq, 2)  #find where z index does not equal row index
    z_neq_c =T.unsqueeze(I_neq, 1)  #find where z index does not equal column index
    r_neq_c = T.unsqueeze(I_neq, 0)  #find where row index does not equal column index
                                
    distinct_mask = T.logical_and(T.logical_and(z_neq_r, z_neq_c), r_neq_c)  #get mask showing true for triplets with 2 unique samples


    # Check if labels[i] == labels[j] and labels[i] != labels[k]
    label_eq = T.eq(T.unsqueeze(labels, 0), T.unsqueeze(labels, 1))  #get 2d matric of showing location of equal labels
    z_eq_r = T.unsqueeze(label_eq, 2)  #convert to column matrices
    z_eq_c = T.unsqueeze(label_eq, 1)  #convert to row matrices

    label_mask = T.logical_and(z_eq_r, T.logical_not(z_eq_c))  #create 3d mask of triplets with valid label combinations

    # Combine the two masks
    mask = T.logical_and(distinct_mask, label_mask)

    return mask


def _get_anchor_positive_triplet_mask(labels, device = 'cpu'):
    """Return a 2D mask where mask[a, p] is True iff a and p are distinct and have same label.
    Args:
        labels: tf.int32 `Tensor` with shape [batch_size]
    Returns:
        mask: tf.bool `Tensor` with shape [batch_size, batch_size]
    """
    # Check that i and j are distinct
    I_eq = T.eye(labels.shape[0], dtype = T.bool).to(device)  #get matrix where indices are equal (BxB boolean identity matrix)
    I_neq = T.logical_not(I_eq)  #invert matrix to find where indices are not equal 

    # Check if labels[i] == labels[j]
    # Uses broadcasting where the 1st argument has shape (1, batch_size) and the 2nd (batch_size, 1)
    labels_eq = T.eq(T.unsqueeze(labels, 0), T.unsqueeze(labels, 1))

    # Combine the two masks
    mask = T.logical_and(I_neq, labels_eq)

    return mask


def _get_anchor_negative_triplet_mask(labels):
    """Return a 2D mask where mask[a, n] is True iff a and n have distinct labels.
    Args:
        labels: tf.int32 `Tensor` with shape [batch_size]
    Returns:
        mask: tf.bool `Tensor` with shape [batch_size, batch_size]
    """
    # Check if labels[i] != labels[k]
    # Uses broadcasting where the 1st argument has shape (1, batch_size) and the 2nd (batch_size, 1)
    labels_eq = T.eq(T.unsqueeze(labels, 0), T.unsqueeze(labels, 1))

    mask = T.logical_not(labels_eq)

    return mask


#-------------------------- Online Triplet Mining -----------------------------

#Online Triplet mining loss using batch all strategy
class Batch_All_Triplet_Loss(T.nn.Module):
  def __init__(self, m= 1.0, metric_matrix = distance_matrix, squared = True, device = 'cpu'):
    super(Batch_All_Triplet_Loss, self).__init__()  # pre 3.3 syntax
    
    self.m = m  # margin or radius
    self.metric_matrix = metric_matrix
    self.squared = squared
    self.fraction_positive_triplets = 0
    self.device = device
    #self.class_hist = []

  def get_fraction_pos(self):
     return self.fraction_positive_triplets
  
    
  def get_class_hist(self):
      return self.class_hist
      
  def forward(self, embeddings, target):
    """Build the triplet loss over a batch of embeddings.

    We generate all the valid triplets and average the loss over the positive ones.

    Args:
        labels: labels of the batch, of size (batch_size,)
        embeddings: tensor of shape (batch_size, embed_dim)
        margin: margin for triplet loss
        squared: Boolean. If true, output is the pairwise squared euclidean distance matrix.
                 If false, output is the pairwise euclidean distance matrix.

    Returns:
        triplet_loss: scalar tensor containing the triplet loss
    """
    class_counts = [0 for x in range(5)]
    
    # Get the pairwise distance matrix
    d_matrix = self.metric_matrix(embeddings, squared = self.squared )
    
    #compute BxBxB distance matrices for anchor to positve and anchor to negative samples
    d_ap_matrix = T.unsqueeze(d_matrix, 2) 
    d_an_matrix = T.unsqueeze(d_matrix, 1)
    #return d_matrix
    # Compute BxBxB Triplet loss matrix
    # triplet_loss[i, j, k] will contain the triplet loss of anchor=i, positive=j, negative=k
    # Uses broadcasting where the 1st argument has shape (B, B, 1)
    # and the 2nd (B, 1, B)
    triplet_loss = d_ap_matrix - d_an_matrix + self.m
    
    # zero loss for invalid triplets
    # (where label(a) != label(p) or label(n) == label(a) or a == p)
    
    mask = _get_triplet_mask(target, device= self.device).float()  #get mask for valid triplets
    
    triplet_loss = mask * triplet_loss  #apply mask to triplet loss matrix
    
    triplet_loss = T.clamp(triplet_loss, min=0.0)  #remove negative losses (easy triplets)
    
    '''
    indicies = T.nonzero(triplet_loss)  
    unique = np.array([T.unique(i).detach().numpy() for i in indicies]).flatten()
    indicies, counts = np.unique(unique, return_counts = True)
    
    for c, i in enumerate(indicies):
        class_counts[target[i]] += counts[c]
    
    self.class_hist.append(class_counts)
    '''
    # Count number of positive triplets (where triplet_loss > 0)
    valid_triplets = T.greater(triplet_loss,1e-16).float()
    num_positive_triplets = T.sum(valid_triplets)
    num_valid_triplets = T.sum(mask)
    self.fraction_positive_triplets = num_positive_triplets / (num_valid_triplets + 1e-16)

    #print(num_positive_triplets)
    #print(num_valid_triplets)

    # Get final mean triplet loss over the positive valid triplets
    triplet_loss = T.sum(triplet_loss) / (num_positive_triplets + 1e-16)


    return triplet_loss

    
  


#semi hard triplet mining
class  Batch_All_Semi_Hard_Triplet_Loss(T.nn.Module):
    def __init__(self, m= 1.0, metric_matrix = distance_matrix, squared = True, device = 'cpu'):
      super(Batch_All_Semi_Hard_Triplet_Loss, self).__init__()  # pre 3.3 syntax
      self.m = m  # margin or radius
      self.metric_matrix = metric_matrix
      self.squared = squared
      self.fraction_positive_triplets = 0
      self.device = 'cpu'
    
    def forward(self, embeddings, target):
        """Build the triplet loss over a batch of embeddings for semi hard triplets.
        For each anchor, we get the hardest positive and hardest negative to form a triplet.
        Args:
            labels: labels of the batch, of size (batch_size,)
            embeddings: tensor of shape (batch_size, embed_dim)
            margin: margin for triplet loss
            squared: Boolean. If true, output is the pairwise squared euclidean distance matrix.
                     If false, output is the pairwise euclidean distance matrix.
        Returns:
            triplet_loss: scalar tensor containing the triplet loss
        """
        
        d_matrix = self.metric_matrix(embeddings, squared= self.squared)
        
        mask_triplet = _get_triplet_mask(target).float()  #get mask for valid triplets
        
        
        
        d_ap = T.unsqueeze(d_matrix, 2)  # col matrices of anchor positive distances

        d_an = T.unsqueeze(d_matrix, 1)  # row matrices of anchor positive distances

        mask_sh = T.gt(d_an,d_ap).float()  #mask filtering out non semihard triplets
        
        
        #calculate triplet loss
        triplet_loss = d_ap - d_an + self.m  #find triplet loss for all triplets
        triplet_loss = triplet_loss * mask_triplet * mask_sh  #apply masks to filter invalid triplets
        triplet_loss = T.mean(T.clamp(triplet_loss, min=0.0))  #remove negative losses (easy triplets) and find mean loss
        
        
        # Count number of positive triplets (where triplet_loss > 0)
        valid_triplets = T.greater(triplet_loss,1e-16).float()
        num_positive_triplets = T.sum(valid_triplets)
        
        num_valid_triplets = T.sum(mask_triplet)
        self.fraction_positive_triplets = num_positive_triplets / (num_valid_triplets + 1e-16)



        # Get final mean triplet loss over the positive valid triplets
        triplet_loss = T.sum(triplet_loss) / (num_positive_triplets + 1e-16)
        
        
        return triplet_loss  #return triplet loss


    def get_fraction_pos(self):
        return self.fraction_positive_triplets


#semi hard triplet mining
class  Batch_Semi_Hard_Triplet_Loss(T.nn.Module):
    def __init__(self, m= 1.0, metric_matrix = distance_matrix, squared = True, device = 'cpu'):
      super(Batch_Semi_Hard_Triplet_Loss, self).__init__()  # pre 3.3 syntax
      self.m = m  # margin or radius
      self.metric_matrix = metric_matrix
      self.squared = squared
      self.device = device
    
    def forward(self, embeddings, target):
        """Build the triplet loss over a batch of embeddings for semi hard triplets.
        For each anchor, we get the positive and negative combination with the greatest loss where D_an > D_ap
        .
        Args:
            labels: labels of the batch, of size (batch_size,)
            embeddings: tensor of shape (batch_size, embed_dim)
            margin: margin for triplet loss
            squared: Boolean. If true, output is the pairwise squared euclidean distance matrix.
                     If false, output is the pairwise euclidean distance matrix.
        Returns:
            triplet_loss: scalar tensor containing the triplet loss
        """
        
        d_matrix = self.metric_matrix(embeddings, squared= self.squared)
        
        mask_triplet = _get_triplet_mask(target, device = self.device).float()  #get mask for valid triplets
        
        
        
        d_ap = T.unsqueeze(d_matrix, 2)  # col matrices of anchor positive distances

        d_an = T.unsqueeze(d_matrix, 1)  # row matrices of anchor positive distances

        mask_sh = T.gt(d_an,d_ap).float()  #mask filtering out non semihard triplets
        
        
        #calculate triplet loss
        triplet_loss = d_ap - d_an + self.m  #find triplet loss for all triplets
        triplet_loss = triplet_loss * mask_triplet * mask_sh  #apply masks to filter invalid triplets
        triplet_loss = T.amax(triplet_loss, dim=(1, 2))  #get hardest semihard triplet for each anchor
        triplet_loss = T.mean(T.clamp(triplet_loss, min=0.0))  #remove negative losses (easy triplets) and find mean loss

        return triplet_loss  #return triplet loss




#Online Triplet mining loss using batch hard strategy
class Batch_Hard_Triplet_Loss(T.nn.Module):
  def __init__(self, m= 1.0, metric_matrix = distance_matrix, squared = True, device = 'cpu'):
    super(Batch_Hard_Triplet_Loss, self).__init__()  # pre 3.3 syntax
    
    self.m = m  # margin or radius
    self.metric_matrix = metric_matrix
    self.squared = squared
    self.device = device

  def forward(self, embeddings, target):
    """Build the triplet loss over a batch of embeddings.
    For each anchor, we get the hardest positive and hardest negative to form a triplet.
    Args:
        labels: labels of the batch, of size (batch_size,)
        embeddings: tensor of shape (batch_size, embed_dim)
        margin: margin for triplet loss
        squared: Boolean. If true, output is the pairwise squared euclidean distance matrix.
                 If false, output is the pairwise euclidean distance matrix.
    Returns:
        triplet_loss: scalar tensor containing the triplet loss
    """
    
    # Get the pairwise distance matrix
    d_matrix = self.metric_matrix(embeddings, squared= self.squared)


    # Get hardest positive sample for each anchor
    mask_ap = _get_anchor_positive_triplet_mask(target, device = self.device).float()  #get mask for positive samples for each anchor
    
    #print(d_matrix)
    #print(mask_ap)
    d_ap = mask_ap * d_matrix  # zero distance for samples with labels which do not match anchor
    d_hard_p = T.max(d_ap, dim = 1, keepdim = True).values  # get maximum distance between anchor and positive samples, for each anchor
    
                        
    # Get hardest negative sample for each anchor
    mask_an = _get_anchor_negative_triplet_mask(target).float()
    d_max_an = T.max(d_matrix, dim=1, keepdim=True).values  #get max distance betweene each anchor and negative sample
    d_an = d_matrix + (d_max_an * (1.0 - mask_an))  #add max distance to invalid pairs to ensure they are not selected for triplet (cannot 0 as they would be selected)

    d_hard_n = T.min(d_an, dim = 1, keepdim = True).values

    triplet_loss = T.clamp(d_hard_p - d_hard_n + self.m, min = 0.0)  # calculate triplet loss using hardest samples for each anchor

    triplet_loss = T.mean(triplet_loss)  # get final mean triplet loss

    return triplet_loss



#--------------------------- Custom Training Loop -----------------------------
'''
def Triplet_Train(model, optimiser, loss_fn, train_dl, epochs, val_dl = None, ep_log_interval = 1, device = 'cpu'):
    print('Training Model')

    #coolect training and validation metrics for each epoch
    history = {}
    history['loss'] = []
    history['val_loss'] = []    
    
    start_time = time.time()
    
    #start training loop
    for epoch in range(epochs):
        
        train_loss = 0  #initalise train loss
        
        model.train()  #set model to train
        
        #--------------- train and evaluate on training dataset ---------------
        
        for batch_num, batch in enumerate(train_dl):
           
            x_a, x_p, x_n = batch  #get validation batch
            

            emb_a, emb_p, emb_n = model(x_a, x_p, x_n)
            #emb_p = model(x_p)
            #emb_n = model(x_n)  #predict euclidean distances for validation batch
            
            optimiser.zero_grad()  #reset gradient values
            
            loss = loss_fn(emb_a, emb_p, emb_n)  #compute training loss
            
            #apply backpropogation
            loss.backward()         # compute gradients 
            optimiser.step()            # update weights
            
            train_loss += loss.item() * x_a.size(0)  # multiply sample loss by batch size for batch loss
            
        train_loss = train_loss / len(train_dl.dataset)  #find per sample loss
        
        
        #-------------------- evaluate on training dataset --------------------
        model.eval()  #set model to evaluat
        val_loss = 0  #init val loss
        
        if val_dl is not None:
            for batch_num, batch in enumerate(val_dl):
            
                x_a, x_p, x_n = batch  #get validation batch
                
                emb_a, emb_p, emb_n = model(x_a, x_p, x_n)
                #emb_a = model(x_a) 
                #emb_p = model(x_p) 
                #emb_n = model(x_n)  #predict euclidean distances for validation batch
                
                loss = loss_fn(emb_a, emb_p, emb_n)  #compute validation loss
                
                val_loss += loss.item() * x_a.size(0)  # multiply sample loss by batch size for batch loss
                
            val_loss = val_loss / len(val_dl.dataset)
            
        history['loss'].append(train_loss)
        history['val_loss'].append(val_loss)
        
        if epoch % ep_log_interval == 0 and ep_log_interval != 0:
            train_time = time.time() - start_time
            print(f"epoch = {epoch}  |  train_loss = {train_loss:.6f}  |  val_loss = {val_loss:.6f} |  training_for: {train_time:.2f}" )
    
    
    
    #end model training
    end_time = time.time()
    total_time = end_time - start_time  #find time to trian
    #time_per_epoch = total_time / epochs  #find time per epoch
    
    print(f'\nTraining Complete in: {total_time:.2f} seconds')
    
    return history
'''


from HPO.utils.visualisation import plot_scores
import csv
from ConfigSpace import ConfigurationSpace
from HPO.algorithms.algorithm_utils import train_eval,load
from HPO.workers.load_eval import evaluate
import json
import copy
import random
import time
import pandas as pd
import numpy as np
from HPO.workers.ensemble import EnsembleManager
import sys

def full_eval(SETTINGS):
  accuracy = {}
  #acc_best_single, recall,params = evaluate("{}/{}".format(SETTINGS["PATH"],"configuration.json"))
  accuracy["single_model"] = 0#acc_best_single
  for i in [1,3,5,10]:
    be = EnsembleManager("{}/{}".format(SETTINGS["PATH"],"configuration.json"),SETTINGS["DEVICES"][0])
    be.get_ensemble(i)
    accuracy["ensemble_{}".format(i)] = be.evaluate(2)
    
  # convert dictionary to dataframe
  df = pd.DataFrame(accuracy, index=[0])

  # save to csv
  df.to_csv('{}/test_results.csv'.format(SETTINGS["PATH"]), index=False)

class model_ts:
  def __init__(self,acc ,recall , config,SETTINGS):
    self.SETTINGS = SETTINGS
    self.ID = config["ID"]
    self.mean_acc = acc
    self.config = config
    self.offspring = 0
    self.evals = 1

  def load_scores(self):
    path = "{}/{}/{}-bin.npy".format(self.SETTINGS["PATH"],"metrics",self.ID)
    x = np.load(path)
    self.a = sum(x) + 1
    self.b = len(x) - sum(x) + 1
    self.evals = len(x)/1500

  def sample(self):
    self.load_scores()
    return np.random.beta(self.a,self.b)

  def sample_mu(self):
    self.load_scores()
    return np.mean(np.random.beta(self.a,self.b,1000))

  def get_config(self):
    self.offspring += 1 
    return self.config

  def get_ratio(self):
    return self.offspring / self.evals



class model:
  def __init__(self,acc ,recall , config,SETTINGS):
    self.SETTINGS = SETTINGS
    self.ID = config["ID"]
    self.mean_acc = acc
    self.config = config
    self.offspring = 0
    self.evals = 1
    self.record_evals = 0
    self.cool = False
    self.flag_need_reload = True
  def load_scores(self):
    path = "{}/{}/{}".format(self.SETTINGS["PATH"],"metrics",self.ID)
    try:
      self.df = pd.read_csv(path)
      self.mu = self.df["accuracy"].mean()
      self.sigma = self.df["accuracy"].std()
      self.evals = len(self.df["accuracy"])
    except:
      self.mu = self.mean_acc

  def sample(self):
    if self.flag_need_reload:
      self.load_scores()
      if not self.cooldown():
        self.flag_need_reload = False
    return self.mu

  def sample_mu(self):
    if self.flag_need_reload:
      self.load_scores()
      if not self.cooldown():
        self.flag_need_reload = False
    return self.mu

  def get_config(self):
    self.offspring += 1 
    return self.config

  def get_ratio(self):
    return self.offspring / self.evals
  def set_cooldown(self):
    self.record_evals = self.evals
    self.flag_need_reload = True
  def cooldown(self):
    return self.record_evals == self.evals

def main(worker, configspace : ConfigurationSpace, json_config):

  #INITIALISATION
  M_FORMAT = model
  with open(json_config) as f:
    data = json.load(f)
    dataset_name = data["WORKER_CONFIG"]["DATASET_CONFIG"]["NAME"]
    SETTINGS = data["SEARCH_CONFIG"]
  TIME_SINCE_IMPROVE = 0
  EARLY_STOP = SETTINGS["EARLY_STOP"]
  START_TIME = time.time()
  RUNTIME = SETTINGS["RUNTIME"]
  last_print = time.time()

  train = train_eval( worker , json_config)




  if SETTINGS["RESUME"]:
    data = load(data["EXPERIMENT_NAME"])
    history.extend([M_FORMAT(s,r,p,SETTINGS) for s,r,p in zip( data["scores"] ,data["recall"] , data["config"])])
    history_scores = data["scores"]
    history_conf = data["config"]
  else:
    configs = configspace.sample_configuration(SETTINGS["INITIAL_POPULATION_SIZE"])
    scores , recall , pop= train.init_async(configs)
    
    history = [M_FORMAT(s,r,p,SETTINGS) for s,r,p in zip(scores ,recall , pop)]

    last_mean_best = None
    iteration = 0
  
  #BEGIN MAIN LOOP
  while iteration < SETTINGS["TOTAL_EVALUATIONS"]:
    scores ,recall , pop = [], [], [] 

    #WAIT FOR NEW RESULTS
    while len(scores) == 0:
      time.sleep(0.05)
      scores ,recall , pop = train.get_async()

    

    history.extend([M_FORMAT(s,r,p,SETTINGS) for s,r,p in zip(scores ,recall , pop)])

    
    #Thompson Sampling
    max_index = np.argmax([i.sample() for i in history])
    mean_best = max([i.sample_mu() for i in history])
    

    while train.config_queue.qsize() < SETTINGS["CORES"]:
      if history[max_index].get_ratio() > 4 and not history[max_index].cooldown():
        history[max_index].set_cooldown()
        train.update_async(history[max_index].get_config())
      else:
        train.update_async(configspace.mutate_graph(history[max_index].get_config(),2))
    if iteration % 10 ==0:
      print("[{}] Generation: {}".format(dataset_name,iteration), " -- Best (Mean) Score: {}".format(mean_best))
    if time.time() > START_TIME + RUNTIME:
      print("Reached Total Alloted Time: {}".format(RUNTIME))
      break
    elif (time.time() - last_print) > 30:
      print("TIME LEFT: {} SECONDS".format((START_TIME + RUNTIME) - time.time() ))
      last_print = time.time()

    if mean_best == last_mean_best:
      TIME_SINCE_IMPROVE+=len(scores)
    else:
      TIME_SINCE_IMPROVE = 0
    last_mean_best = copy.copy(mean_best)
    if TIME_SINCE_IMPROVE > EARLY_STOP:
      break
    iteration+= 1

  train.kill_workers()
  full_eval(SETTINGS)
if __name__ == "__main__":
  with open(sys.argv[1]) as f:
    DATA = json.load(f)
    full_eval(DATA["SEARCH_CONFIG"])


import random
import os 
import numpy as np
import torch
import copy
from HPO.searchspaces.DARTS_config import init_config
from torch.utils.data import Dataset
from HPO.algorithms.algorithm_utils import train_eval
DEBUG_CONST = True
if DEBUG_CONST:
  random.seed(1)
  np.random.seed(1)
  torch.random.manual_seed(1) 

class Labels:
  def __init__(self,clean, classes , samples):
    self.values = clean
    self.is_clean = {x:x in clean for x in range(samples)}
    self.is_predicted = {x:False for x in range(samples)}
    self.is_valid = {x:(self.is_clean[x] or self.is_predicted[x]) for x in range(samples)} #List of indexes with any kind of label
    self.samples = samples
    self.classes = classes
    #self.generate_random()

  def __getitem__(self,index):
    return self.values[index]

  def generate_random(self):
    empty = [ x for x in range(self.samples) if x not in set(list(self.values.keys()))]
    for index in empty:
      self.values[index] = random.choice(list(range(self.classes)))

  def get_valid_samples(self):
    return self.is_valid


class SampleSet(Dataset):
  def __init__(self,mapping):
    self.mapping = mapping
    self.n_samples = len(self.mapping)-1
    self.parent = None 
  def set_main_dataset(self,dataset):
    self.parent = dataset
  def __getitem__(self, index):
    index = self.mapping[index]
    return self.parent[index]
  def __len__(self):
    return self.n_samples
  def get_n_classes(self):
    return self.parent.get_n_classes()
  def get_n_samples(self):
    return self.n_samples
  def get_n_features(self):
    return self.parent.get_n_features()
  def __len__(self):
    return self.n_samples



class Generic(Dataset):
  def __init__(self, x_samples, n_classes , n_features, clean_labels):
    """
    x_samples : array-like of all samples
    clean_labels : diction of confirmed correct labels {index_of_sample : label_value}
    noisy_labels : diction of non confirmed labels {index_of_sample : label_value}

    """
    self.n_features = n_features
    self.n_classes = n_classes
    self.n_samples = len(x_samples)
    self.x_samples = x_samples 
    self.labels = clean_labels
  def __getitem__(self, index):
    x = self.x_samples[index]
    y = self.labels[index]
    return x , y
  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples
  def get_n_features(self):
    return self.n_features
  def __len__(self):
    return self.n_samples

"""
class SampleSet(Generic):
  def __init__(self,x_samples,n_classes,n_features,labels):
    self.n_samples = x_samples.shape
    super().__init__(x_samples, n_classes , n_features, labels)
"""
class BootSet(Generic):
  def __init__(self, x_samples, n_classes , n_features, clean_labels = None , noisy_labels = None ):
    """
    x_samples : array-like of all samples
    clean_labels : diction of confirmed correct labels {index_of_sample : label_value}
    noisy_labels : diction of non confirmed labels {index_of_sample : label_value}
    """
    self.n_samples = len(x_samples)
    print("Number of samples: {}".format(self.n_samples))
    self.x_samples = x_samples 
    self.labels = Labels(clean_labels, n_classes,self.n_samples)
    self.n_features = n_features
    self.n_classes = n_classes

  def generate_training_sample(self,size):
    mapping,indexs = self.generate_mapping(size)
    return SampleSet(mapping)

  def generate_mapping(self, size):
    is_valid = self.labels.get_valid_samples()
    mapping  = {}
    sub_sample_indexs = []
    samples = list(range(self.n_samples-1))
    index = 0
    while len(mapping) < size:
      sample = samples.pop(random.randint(0,len(samples)-1))
      if is_valid[sample]:
        mapping[index] = sample
        sub_sample_indexs.append(sample)
        index+=1
    return mapping,sub_sample_indexs
      
def split_x_y(files : list, path : str):
    x_samples = {}
    clean_labels = {}
    for index ,i in enumerate(files):
      data = np.reshape(np.load(path+i),(53,-1))
      x = data[1:,:]
      y = data[0,0]
      x_samples[index] = x 
      clean_labels[index]=y
    return x_samples , clean_labels
    

def load_test():

    path = "/home/snaags/scripts/datasets/TEPS/split/"
    files = os.listdir(path)
    filtr = "testing"
    files_test = [ name for name in files if filtr in name]    
    x_validation , y_validation = split_x_y(files_test,path)
    #Random subset of keys and samples 
    
    n_features = 52
    n_classes = 21
    return Generic(x_validation,  n_classes , n_features, y_validation) 

def load_train(num_clean = 10000,total_samples = 10000):
    path = "/home/snaags/scripts/datasets/TEPS/split/"
    files = os.listdir(path)
    filtr = "training"
    files_train = [ name for name in files if filtr in name]    
    x_samples , clean_labels = split_x_y(files_train, path)
    #Random subset of keys and samples 
    if total_samples != None:
      subset_keys = np.random.choice( a = list(x_samples.keys()) , size = total_samples ,replace = False)
      #x_samples = {key:torch.from_numpy(x_samples[key]) for key in subset_keys}
      #clean_labels = {key:torch.from_numpy(clean_labels[key]) for key in subset_keys}
      x_samples = {key:x_samples[key] for key in subset_keys}
      clean_labels = {key:clean_labels[key] for key in subset_keys}
      samples = {}
      labels = {}
      for new_index, (label,sample) in enumerate(zip(clean_labels,x_samples)):
        samples[new_index] = x_samples[sample]
        labels[new_index] = clean_labels[sample]
    #Random subset keys for testing
    clean_labels_subset_keys = np.random.choice( a = list(labels.keys()) , size = num_clean ,replace = False)
    clean_labels_subset = {key:labels[key] for key in clean_labels_subset_keys}

    
    n_features = 52
    n_classes = 21
    return BootSet(samples, n_classes, n_features, clean_labels_subset)

def compute_size():
  #Place holder
  return 500

global test_data
test_data = load_test()
global bootloader
bootloader = load_train()
def main(worker):
    
  #Settings
  N_ITERATIONS = 100
  CORES = 1
  BATCH = 24
  ##Set Up  
  cs = init_config()
  train = train_eval( worker, CORES, filename = "bootloader.csv", handle_dataset = True)
  

  for i in range(N_ITERATIONS):
    ##Iteration
    configs = []
    training_samples = []
    for _ in range(BATCH):
      size = compute_size()
      configs.append( cs.sample_configuration())
      training_samples.append([ bootloader.generate_training_sample(size)])
      
    acc , recall , config = train.eval( population = configs, datasets =training_samples )
    print(acc, recall)
    
if __name__ == "__main__":
  from HPO.workers.boot_worker import compute as worker
  
  main(worker)
  











import torch
import numpy

def freeze_normal_cells(model):

  for cell in model.normal_cells:
    for mList in cell.compute_order:
      for operation in mList.compute:
        for param in operation.parameters():
          param.requires_grad = False
  return model
          
def freeze_reduction_cells(model):
  for cell in model.reduction_cells:
    for mList in cell.compute_order:
      for operation in mList.compute:
        for param in operation.parameters():
          param.requires_grad = False

  return model

def freeze_all_cells( model ):
  model = freeze_reduction_cells( model )
  model = freeze_normal_cells( model )
  return model

def freeze_FCN( model ):
  for l in model.layers:
     for param in l.parameters():
          param.requires_grad = False

  return model

def freeze_resnet( model ):
  for l in model.layer_1:
     for param in l.parameters():
          param.requires_grad = False
  for l in model.layer_2:
     for param in l.parameters():
          param.requires_grad = False
  
  for l in model.layer_3:
     for param in l.parameters():
          param.requires_grad = False

  return model



from collections import namedtuple

Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')

PRIMITIVES = [
    'none',
    'max_pool_3x3',
    'avg_pool_3x3',
    'skip_connect',
    'sep_conv_3x3',
    'sep_conv_5x5',
    'dil_conv_3x3',
    'dil_conv_5x5'
]

NASNet = Genotype(
  normal = [
    ('sep_conv_5x5', 1),
    ('sep_conv_3x3', 0),
    ('sep_conv_5x5', 0),
    ('sep_conv_3x3', 0),
    ('avg_pool_3x3', 1),
    ('skip_connect', 0),
    ('avg_pool_3x3', 0),
    ('avg_pool_3x3', 0),
    ('sep_conv_3x3', 1),
    ('skip_connect', 1),
  ],
  normal_concat = [2, 3, 4, 5, 6],
  reduce = [
    ('sep_conv_5x5', 1),
    ('sep_conv_7x7', 0),
    ('max_pool_3x3', 1),
    ('sep_conv_7x7', 0),
    ('avg_pool_3x3', 1),
    ('sep_conv_5x5', 0),
    ('skip_connect', 3),
    ('avg_pool_3x3', 2),
    ('sep_conv_3x3', 2),
    ('max_pool_3x3', 1),
  ],
  reduce_concat = [4, 5, 6],
)
    
AmoebaNet = Genotype(
  normal = [
    ('avg_pool_3x3', 0),
    ('max_pool_3x3', 1),
    ('sep_conv_3x3', 0),
    ('sep_conv_5x5', 2),
    ('sep_conv_3x3', 0),
    ('avg_pool_3x3', 3),
    ('sep_conv_3x3', 1),
    ('skip_connect', 1),
    ('skip_connect', 0),
    ('avg_pool_3x3', 1),
    ],
  normal_concat = [4, 5, 6],
  reduce = [
    ('avg_pool_3x3', 0),
    ('sep_conv_3x3', 1),
    ('max_pool_3x3', 0),
    ('sep_conv_7x7', 2),
    ('sep_conv_7x7', 0),
    ('avg_pool_3x3', 1),
    ('max_pool_3x3', 0),
    ('max_pool_3x3', 1),
    ('conv_7x1_1x7', 0),
    ('sep_conv_3x3', 5),
  ],
  reduce_concat = [3, 4, 6]
)

DARTS_V1 = Genotype(normal=[('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('skip_connect', 0), ('sep_conv_3x3', 1), ('skip_connect', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('skip_connect', 2)], normal_concat=[2, 3, 4, 5], reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 0), ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('avg_pool_3x3', 0)], reduce_concat=[2, 3, 4, 5])
DARTS_V2 = Genotype(normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 1), ('skip_connect', 0), ('skip_connect', 0), ('dil_conv_3x3', 2)], normal_concat=[2, 3, 4, 5], reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 1), ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('max_pool_3x3', 1)], reduce_concat=[2, 3, 4, 5])

DARTS = DARTS_V2




import torch
import numpy as np
import math
import torch.nn as nn
import torch.nn.functional as F
from custom_transformers.time_series_positional_encoding import LEEM, TAPE,ERPE
ACTIVATION = {
  "gelu" : lambda C, L: nn.GELU(),
  "relu" : lambda  C, L: nn.ReLU(),
  "none" : lambda  C, L: Identity(),
  "tanh" : lambda  C, L: nn.Tanh(),
  "mha" : lambda  C, L: MHA(C)
}
NORMALISATION = {
  "batch_norm" : lambda C, l : nn.BatchNorm1d(C),  
  "layer_norm" : lambda C, l : nn.LayerNorm([C,l]),
  "none" : lambda C, l : Identity()

}

COMBINE = {
  "CONCAT" : lambda : CONCAT(),
  "ADD" : lambda : ADD()
}

OPS = {
  "none" : lambda C : Zero(),
  "gelu" : lambda C,kernel, stride,dil, affine: nn.GELU(),
  "relu" : lambda : nn.ReLU(),
  "batch_norm" : lambda C : nn.BatchNorm1d(C),
  "skip_connect" : lambda C : Identity(),
  "MHA_2" : lambda C : MHA(C, 2),
  "MHA_4" : lambda C : MHA(C, 4),
  "MHA_8" : lambda C : MHA(C, 8),
  "L_ERPE_4" : lambda C ,L: ERPE(C,L, 4),
  "L_TAPE_4" : lambda C ,L: TAPE(C,L, 4),
  "L_LEEM_4" : lambda C ,L: LEEM(C,L, 4),

  "SE_1" : lambda C : SE(C, 1),
  "SE_2" : lambda C : SE(C, 2),
  "SE_4" : lambda C : SE(C, 4),
  "lstm" :lambda C : LSTM(C, False),
  "lstm_bi" :lambda C : LSTM(C, True),
  "depth_lstm" :lambda C : DepthLSTM(C, False),
  "depth_lstm_bi" :lambda C : DepthLSTM(C, True),
  "resample_channels" : lambda c_in,c_out: ResampleChannels(c_in,c_out),
  "downsample_resolution" : lambda c_in,stride: DownsampleResolution(c_in,stride),
  "point_conv_1" :  lambda C: PointConv(C, stride = 1,dilation = 1,padding = 1*(3 - 1)//2, affine = True),
  #DEPTHWISE CONVOLUTIONS
  "depth_conv_3_1" :  lambda C: DepthConv(C,3, stride = 1,dilation = 1,padding = 1*(3 - 1)//2, affine = True),
  "depth_conv_5_1" :  lambda C: DepthConv(C,5, stride = 1,dilation = 1,padding = 1*(5 - 1)//2, affine = True),
  "depth_conv_7_1" :  lambda C: DepthConv(C,7, stride = 1,dilation = 1,padding = 1*(7 - 1)//2, affine = True),

  "depth_conv_3_2" :  lambda C: DepthConv(C,3, stride = 1,dilation = 2,padding = 2*(3 - 1)//2, affine = True),
  "depth_conv_5_2" :  lambda C: DepthConv(C,5, stride = 1,dilation = 2,padding = 2*(5 - 1)//2, affine = True),
  "depth_conv_7_2" :  lambda C: DepthConv(C,7, stride = 1,dilation = 2,padding = 2*(7 - 1)//2, affine = True),
  "depth_conv_3_4" :  lambda C: DepthConv(C,3, stride = 1,dilation = 4,padding = 4*(3 - 1)//2, affine = True),
  "depth_conv_5_4" :  lambda C: DepthConv(C,5, stride = 1,dilation = 4,padding = 4*(5 - 1)//2, affine = True),
  "depth_conv_7_4" :  lambda C: DepthConv(C,7, stride = 1,dilation = 4,padding = 4*(7 - 1)//2, affine = True),
  "depth_conv_3_8" :  lambda C: DepthConv(C,3, stride = 1,dilation = 8,padding = 8*(3 - 1)//2, affine = True),
  "depth_conv_5_8" :  lambda C: DepthConv(C,5, stride = 1,dilation = 8,padding = 8*(5 - 1)//2, affine = True),
  "depth_conv_7_8" :  lambda C: DepthConv(C,7, stride = 1,dilation = 8,padding = 8*(7 - 1)//2, affine = True),
  "depth_conv_3_16" :  lambda C: DepthConv(C,3, stride = 1,dilation = 16,padding = 16*(3 - 1)//2, affine = True),
  "depth_conv_5_16" :  lambda C: DepthConv(C,5, stride = 1,dilation = 16,padding = 16*(5 - 1)//2, affine = True),
  "depth_conv_7_16" :  lambda C: DepthConv(C,7, stride = 1,dilation = 16,padding = 16*(7 - 1)//2, affine = True),
  "depth_conv_3_32" :  lambda C: DepthConv(C,3, stride = 1,dilation = 32,padding = 32*(3 - 1)//2, affine = True),
  "depth_conv_5_32" :  lambda C: DepthConv(C,5, stride = 1,dilation = 32,padding = 32*(5 - 1)//2, affine = True),
  "depth_conv_7_32" :  lambda C: DepthConv(C,7, stride = 1,dilation = 32,padding = 32*(7 - 1)//2, affine = True),
  "depth_conv_15_1" :  lambda C: DepthConv(C,15, stride = 1,dilation = 1,padding = 1*(15 - 1)//2, affine = True),
  "depth_conv_31_1" :  lambda C: DepthConv(C,31, stride = 1,dilation = 1,padding = 1*(31 - 1)//2, affine = True),
  "depth_conv_63_1" :  lambda C: DepthConv(C,63, stride = 1,dilation = 1,padding = 1*(63 - 1)//2, affine = True),
  "depth_conv_15_2" :  lambda C: DepthConv(C,15, stride = 1,dilation = 2,padding = 2*(15 - 1)//2, affine = True),
  "depth_conv_31_2" :  lambda C: DepthConv(C,31, stride = 1,dilation = 2,padding = 2*(31 - 1)//2, affine = True),
  "depth_conv_63_2" :  lambda C: DepthConv(C,63, stride = 1,dilation = 2,padding = 2*(63 - 1)//2, affine = True),
  "depth_conv_15_4" :  lambda C: DepthConv(C,15, stride = 1,dilation = 4,padding = 4*(15 - 1)//2, affine = True),
  "depth_conv_31_4" :  lambda C: DepthConv(C,31, stride = 1,dilation = 4,padding = 4*(31 - 1)//2, affine = True),
  "depth_conv_63_4" :  lambda C: DepthConv(C,63, stride = 1,dilation = 4,padding = 4*(63 - 1)//2, affine = True),
  "depth_conv_15_8" :  lambda C: DepthConv(C,15, stride = 1,dilation = 8,padding = 8*(15 - 1)//2, affine = True),
  "depth_conv_31_8" :  lambda C: DepthConv(C,31, stride = 1,dilation = 8,padding = 8*(31 - 1)//2, affine = True),
  "depth_conv_63_8" :  lambda C: DepthConv(C,63, stride = 1,dilation = 8,padding = 8*(63 - 1)//2, affine = True),
  "depth_conv_15_16" :  lambda C: DepthConv(C,15, stride = 1,dilation = 16,padding =16*(15 - 1)//2, affine = True),
  "depth_conv_31_16" :  lambda C: DepthConv(C,31, stride = 1,dilation = 16,padding = 8*(31 - 1)//2, affine = True),
  "depth_conv_63_16" :  lambda C: DepthConv(C,63, stride = 1,dilation = 16,padding = 8*(31 - 1)//2, affine = True),
  "depth_conv_15_32" :  lambda C: DepthConv(C,15, stride = 1,dilation = 32,padding =16*(15 - 1)//2, affine = True),
  "depth_conv_31_32" :  lambda C: DepthConv(C,31, stride = 1,dilation = 32,padding =16*(15 - 1)//2, affine = True),
  "depth_conv_63_32" :  lambda C: DepthConv(C,63, stride = 1,dilation = 32,padding =32*(63 - 1)//2, affine = True),
  #AVERAGE POOLING OPERATIONS
  "avg_pool_3_1" :  lambda C: nn.AvgPool1d(3,stride = 1, padding = 1*(3 - 1)//2,count_include_pad=False),
  "avg_pool_5_1" :  lambda C: nn.AvgPool1d(5,stride = 1,padding = 2,count_include_pad=False),
  "avg_pool_7_1" :  lambda C: nn.AvgPool1d(7,stride = 1,padding = 3,count_include_pad=False),
  "avg_pool_15_1" :  lambda C: nn.AvgPool1d(15,stride = 1,padding = 7,count_include_pad=False),
  "avg_pool_31_1" :  lambda C: nn.AvgPool1d(31,stride = 1,padding = 15,count_include_pad=False),
  "avg_pool_63_1" :  lambda C: nn.AvgPool1d(63,stride = 1,padding = 31,count_include_pad=False),
  #MAXPOOLING OPERATIONS
  "max_pool_3_1" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 1,padding = 1),
  "max_pool_5_1" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 1,padding = 2),
  "max_pool_7_1" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 1,padding = 3),
  "max_pool_15_1" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 1,padding = 7),
  "max_pool_31_1" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 1,padding = 15),
  "max_pool_63_1" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 1,padding = 31),
  "max_pool_3_2" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 2,padding = 2*(3 - 1)//2),
  "max_pool_5_2" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 2,padding = 2*(5 - 1)//2),
  "max_pool_7_2" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 2,padding = 2*(7 - 1)//2),
  "max_pool_3_4" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 4,padding = 4*(3 - 1)//2),
  "max_pool_5_4" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 4,padding = 4*(5 - 1)//2),
  "max_pool_7_4" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 4,padding = 4*(7 - 1)//2),
  "max_pool_3_8" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 8,padding = 8*(3 - 1)//2),
  "max_pool_5_8" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 8,padding = 8*(5 - 1)//2),
  "max_pool_7_8" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 8,padding = 8*(7 - 1)//2),
  "max_pool_3_16" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 16,padding = 16*(3 - 1)//2),
  "max_pool_5_16" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 16,padding = 16*(5 - 1)//2),
  "max_pool_7_16" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 16,padding = 16*(7 - 1)//2),
  "max_pool_3_32" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 32,padding = 32*(3 - 1)//2),
  "max_pool_5_32" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 32,padding = 32*(5 - 1)//2),
  "max_pool_7_32" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 32,padding = 32*(7 - 1)//2),
  "max_pool_15_2" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 2,padding = 2*(15 - 1)//2),
  "max_pool_31_2" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 2,padding = 2*(31 - 1)//2),
  "max_pool_63_2" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 2,padding = 2*(63 - 1)//2),
  "max_pool_15_4" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 4,padding = 4*(15 - 1)//2),
  "max_pool_31_4" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 4,padding = 4*(31 - 1)//2),
  "max_pool_63_4" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 4,padding = 4*(63 - 1)//2),
  "max_pool_15_8" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 8,padding = 8*(15 - 1)//2),
  "max_pool_31_8" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 8,padding = 8*(31 - 1)//2),
  "max_pool_63_8" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 8,padding = 8*(63 - 1)//2),
  "max_pool_15_16" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 16,padding =16*(15 - 1)//2),
  "max_pool_31_16" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 16,padding =16*(31 - 1)//2),
  "max_pool_63_16" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 16,padding =16*(63 - 1)//2),
  "max_pool_15_32" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 32,padding =32*(15 - 1)//2),
  "max_pool_31_32" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 32,padding =32*(31 - 1)//2),
  "max_pool_63_32" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 32,padding =32*(63 - 1)//2),

  'sep_conv_3_BR' : lambda C : SepConv(C, C, 3, 1, 1, affine=1),
  'sep_conv_5_BR' : lambda C : SepConv(C, C, 5, 1, 2, affine=1),
  'sep_conv_7_BR' : lambda C : SepConv(C, C, 7, 1, 3, affine=1),
  'sep_conv_15_BR' : lambda C, : SepConv(C, C, 15, 1, 7, affine=1),
  'sep_conv_30_BR' : lambda C: SepConv(C, C, 27, 1, 13, affine=1),
  'dil_conv_3_BR' : lambda C: DilConv(C, C, 3, 1, 2, 2, affine=1),
  'dil_conv_5_BR' : lambda C: DilConv(C, C, 5, 1, 4, 2, affine=1),
  'dil_conv_7_BR' : lambda C: DilConv(C, C, 3, 1, 2, 2, affine=1),

  "depth_conv_3_1_BR" :  lambda C: DepthConvBR(C,3, stride = 1,dilation = 1,padding = 1*(3 - 1)//2, affine = True),
  "depth_conv_5_1_BR" :  lambda C: DepthConvBR(C,5, stride = 1,dilation = 1,padding = 1*(5 - 1)//2, affine = True),
  "depth_conv_7_1_BR" :  lambda C: DepthConvBR(C,7, stride = 1,dilation = 1,padding = 1*(7 - 1)//2, affine = True),
  "depth_conv_3_2_BR" :  lambda C: DepthConvBR(C,3, stride = 1,dilation = 2,padding = 2*(3 - 1)//2, affine = True),
  "depth_conv_5_2_BR" :  lambda C: DepthConvBR(C,5, stride = 1,dilation = 2,padding = 2*(5 - 1)//2, affine = True),
  "depth_conv_7_2_BR" :  lambda C: DepthConvBR(C,7, stride = 1,dilation = 2,padding = 2*(7 - 1)//2, affine = True),
  "point_conv_1_BR" :  lambda C: PointConvBR(C, stride = 1,dilation = 1,padding = 1*(3 - 1)//2, affine = True),
  
  }

"""
class MHA(nn.Module):
  def __init__(self,C):
    super(MHA,self).__init__()
    self.act = nn.MultiheadAttention(C, C, batch_first = True)
  def forward(self,x):
    x = x.swapaxes(1,2)
    return self.act(x,x,x,need_weights=False)[0].swapaxes(1,2)
"""

class Identity(nn.Module):

  def __init__(self):
    super(Identity, self).__init__()

  def forward(self, x):
    return x

class CONCAT(nn.Module):

  def __init__(self):
    super(CONCAT, self).__init__()

  def forward(self, x):
    return torch.cat(x,dim = 1)

class ADD(nn.Module):

  def __init__(self):
    super(ADD, self).__init__()

  def forward(self, x):
    x = torch.stack(x, dim=0)
    return torch.sum(x,dim = 0)

class Zero(nn.Module):

  def __init__(self):
    super(Zero, self).__init__()

  def forward(self, x):
    return x.mul(0.)


class GRU(nn.Module):
  def __init__(self, c):
    super(GRU, self).__init__()
    self.gru = nn.GRU(
        input_size=c,
        hidden_size=c,
        num_layers=1,
        batch_first=True,
        bias = False
        )

class LSTM(nn.Module):
  def __init__(self, c,bidirectional = False):
    super(LSTM, self).__init__()
    self.lstm = nn.LSTM(
        input_size=c,
        hidden_size=c,
        num_layers=1,
        batch_first=True,
        bias = False
        )
  def forward(self,x):
    x = x.permute(0,2,1)
    x, (_1,_2) = self.lstm(x)
    x = x.permute(0,2,1)
    return x

class DepthLSTM(nn.Module):
  def __init__(self, c,bidirectional = False):
    super(DepthLSTM, self).__init__()
    self.lstms = nn.ModuleList(
      [
            nn.LSTM(input_size=1, hidden_size=1, num_layers=1, batch_first=True,bias = False,bidirectional = bidirectional)
            for _ in range(c)
        ])

  def forward(self,x):
    outputs = []
    for i, lstm in enumerate(self.lstms):
        input_channel = x[:, i, :].unsqueeze(-1)  # Select the ith feature and add feature dimension
        output, (_1,_2) = lstm(input_channel)
        outputs.append(output)
    outputs = torch.cat(outputs, dim=-1)  # Concatenate outputs along the feature dimension
    return outputs.permute(0,2,1)
class DownsampleResolution(nn.Module):
  """
  This operation uses a depthwise convolution to reduce the signal resolution when needed.
  Given a stride of 2 its a kernel of 2 etc. 
  """
  def __init__(self, c_in,stride):
    super(DownsampleResolution,self).__init__()
    self.stride = stride
    self.conv1 = nn.Conv1d(c_in,c_in, kernel_size = stride, stride = stride, bias = False, groups = c_in)
  def forward(self,x):
      return self.conv1(x)

class ResampleChannels(nn.Module):
  """
  This operation uses a pointwise convolution to change the number of channels when needed.
  """
  def __init__(self,c_in,c_out):
    super(ResampleChannels, self).__init__()
    self.conv1 = nn.Conv1d(c_in,c_out, kernel_size = 1, stride = 1, bias = False)
  def forward(self,x):
    return self.conv1(x)

class GELU(nn.Module):
  def __init__(self):
    super(GELU,self).__init__()
    self.act = nn.GELU()
  def forward(self,x):
    return self.act(x)

class SEMIX(nn.Module):
  def __init__(self, C_in,C_out,r =2 ,stride =1,affine = True ):
    super(SEMIX,self).__init__()
    #print("Building Squeeze Excite with input {} and output: {}".format(C_in,C_out))
    self.GP = nn.AdaptiveAvgPool1d(1)
    self.fc1 = nn.Linear(C_in, C_in//2, bias = False)
    self.act = nn.GELU()
    self.fc2 = nn.Linear(C_in//2, C_out ,bias = False)
    self.sig = nn.Sigmoid()
    self.stride = stride
  def forward(self,x1,x2):
    #Squeeze
    y = self.GP(x2).squeeze()# [Batch,C]
    #torch.mean(x,axis = 2)  
    y = self.fc1(y)
    y = self.act(y)
    y = self.sig(y).unsqueeze(dim = 2)
    return x1* y.expand_as(x1)


class LayerNorm(nn.Module):
  def __init__(self,c,affine):
    super(LayerNorm,self).__init__()
    self.norm = nn.LayerNorm(c, affine=affine)
  def forward(self,x):
    return self.norm(x)

class BatchNorm(nn.Module):
  def __init__(self,c,affine):
    super(BatchNorm,self).__init__()
    self.norm = nn.BatchNorm1d(c, affine=affine)
  def forward(self,x):
    return self.norm(x)

class DepthConv(nn.Module):
  def __init__(self, c, kernel_size, padding,stride,dilation= 1, affine=True):
    super(DepthConv, self).__init__()
    #padding = (kernel_size*stride*dilation)//2
    self.conv = nn.Conv1d(c,c, kernel_size=kernel_size, stride=stride, padding="same", groups=c, bias=False)

  def forward(self, x):
    x =self.conv(x)
    return x

class DepthConvBR(nn.Module):
  def __init__(self, c, kernel_size, padding,stride,dilation= 1, affine=True):
    super(DepthConvBR, self).__init__()
    #padding = (kernel_size*stride*dilation)//2
    self.act =  nn.ReLU()
    self.bn = nn.BatchNorm1d(c, affine=affine)
    self.conv = nn.Conv1d(c,c, kernel_size=kernel_size, stride=stride, padding="same", groups=c, bias=False)

  def forward(self, x):
    x = self.act(x)
    x =self.conv(x)
    x = self.bn(x)
    return x

class PointConv(nn.Module):
  def __init__(self, C, stride,affine=True,dilation = 1,padding = 0 ):
    super(PointConv, self).__init__()
    if stride == 2:
      self.conv = nn.Conv1d(C,C*2, kernel_size=1, dilation = 1,stride=stride, bias=False)
    else:
      self.conv = nn.Conv1d(C,C, kernel_size=1, stride=stride,dilation = 1 , bias=False)

  def forward(self, x):
    x =self.conv(x)
    return x

class PointConvBR(nn.Module):
  def __init__(self, C, stride,affine=True,dilation = 1,padding = 0 ):
    super(PointConvBR, self).__init__()
    if stride == 2:
      self.conv = nn.Conv1d(C,C*2, kernel_size=1, dilation = 1,stride=stride, bias=False)
    else:
      self.conv = nn.Conv1d(C,C, kernel_size=1, stride=stride,dilation = 1 , bias=False)
    self.act =  nn.ReLU()
    self.bn = nn.BatchNorm1d(C, affine=affine)

  def forward(self, x):
    x = self.act(x)
    x =self.conv(x)
    x = self.bn(x)
    return x


class SepConv(nn.Module):
    
  def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
    super(SepConv, self).__init__()
    self.op = nn.Sequential(
      nn.ReLU(inplace=False),
      nn.Conv1d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, groups=C_in, bias=False),
      nn.Conv1d(C_in, C_in, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm1d(C_in, affine=affine),
      nn.ReLU(inplace=False),
      nn.Conv1d(C_in, C_in, kernel_size=kernel_size, stride=1, padding=padding, groups=C_in, bias=False),
      nn.Conv1d(C_in, C_out, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm1d(C_out, affine=affine),
      )

  def forward(self, x):
    return self.op(x)

class DilConv(nn.Module):
    
  def __init__(self, C_in, C_out, kernel_size, stride, padding, dilation, affine=True):
    super(DilConv, self).__init__()
    self.op = nn.Sequential(
      nn.ReLU(inplace=False),
      nn.Conv1d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, groups=C_in, bias=False),
      nn.Conv1d(C_in, C_out, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm1d(C_out, affine=affine),
      )

  def forward(self, x):
    return self.op(x)


class SE(nn.Module):
    def __init__(self, channels, reduction=16):
        super(SE, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool1d(1)
        self.fc = nn.Sequential(
            nn.Linear(channels, channels // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(channels // reduction, channels, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x):
        batch_size, channels, length = x.size()
        y = self.avg_pool(x).view(batch_size, channels)
        y = self.fc(y).view(batch_size, channels, 1)
        return x * y.expand_as(x)


class MHA(nn.Module):
    def __init__(self, d_model, num_heads):
        super(MHA, self).__init__()

        assert d_model % num_heads == 0, "d_model must be divisible by num_heads"
        
        self.d_model = d_model
        self.num_heads = num_heads
        self.head_dim = d_model // num_heads

        self.q_linear = nn.Linear(d_model, d_model)
        self.k_linear = nn.Linear(d_model, d_model)
        self.v_linear = nn.Linear(d_model, d_model)
        self.out_linear = nn.Linear(d_model, d_model)
        self.pe = None

    def forward(self, x):
        batch_size = x.size(0)
        if self.pe == None or self.pe.shape[0] != 16:
          pe = torch.zeros(x.shape)
          pos = torch.arange(x.shape[2]).unsqueeze(1).view(1,-1)
          denominator = torch.exp(torch.arange(0,self.d_model,2)*(-math.log(10000.0)/self.d_model)).unsqueeze(1)
          pe[:,0::2,:] = torch.sin(pos *denominator)
          pe[:,1::2,:] = torch.cos(pos *denominator)
          self.pe = pe.cuda(x.device)
        x += self.pe
        # Calculate q, k, and v
        x = x.permute(0, 2, 1)
        q = self.q_linear(x)
        k = self.k_linear(x)
        v = self.v_linear(x)

        # Split the last dimension into (heads, depth)
        q = q.view(batch_size, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        k = k.view(batch_size, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        v = v.view(batch_size, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)

        # Scaled Dot-Product Attention
        attn_weights = torch.matmul(q, k.transpose(-1, -2)) / torch.sqrt(torch.tensor(self.head_dim, dtype=torch.float))
        attn_weights = torch.softmax(attn_weights, dim=-1)
        attn_output = torch.matmul(attn_weights, v)

        # Concatenate heads and put through final linear layer
        attn_output = attn_output.permute(0, 2, 1, 3).contiguous()
        attn_output = attn_output.view(batch_size, -1, self.d_model)

        output = self.out_linear(attn_output)
        output = output.permute(0, 2, 1)
        return output


class MHA_pt(nn.Module):
    def __init__(self, d_model, num_heads):
        super(MHA, self).__init__()

        assert d_model % num_heads == 0, "d_model must be divisible by num_heads"
        
        self.d_model = d_model
        self.num_heads = num_heads
        self.head_dim = d_model // num_heads

        self.q_linear = nn.Linear(d_model, d_model)
        self.k_linear = nn.Linear(d_model, d_model)
        self.v_linear = nn.Linear(d_model, d_model)
        self.out_linear = nn.Linear(d_model, d_model)
        self.pe = None
        self.mha = nn.MultiheadAttention(d_model, num_heads, dropout=0.0, bias=False, add_bias_kv=False, add_zero_attn=False, kdim=None, vdim=None, batch_first=True)

    def forward(self, x):
        batch_size = x.size(0)
        if self.pe == None or self.pe.shape[0] != 16:
          pe = torch.zeros(x.shape)
          pos = torch.arange(x.shape[2]).unsqueeze(1).view(1,-1)
          denominator = torch.exp(torch.arange(0,self.d_model,2)*(-math.log(10000.0)/self.d_model)).unsqueeze(1)
          pe[:,0::2,:] = torch.sin(pos *denominator)
          pe[:,1::2,:] = torch.cos(pos *denominator)
          self.pe = pe.cuda(x.device)
        x += self.pe
        # Calculate q, k, and v


        x = x.permute(0, 2, 1)
        q = self.q_linear(x)
        k = self.k_linear(x)
        v = self.v_linear(x)

        # Split the last dimension into (heads, depth)
        q = q.view(batch_size, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        k = k.view(batch_size, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        v = v.view(batch_size, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)

        # Scaled Dot-Product Attention
        attn_weights = torch.matmul(q, k.transpose(-1, -2)) / torch.sqrt(torch.tensor(self.head_dim, dtype=torch.float))
        attn_weights = torch.softmax(attn_weights, dim=-1)
        attn_output = torch.matmul(attn_weights, v)

        # Concatenate heads and put through final linear layer
        attn_output = attn_output.permute(0, 2, 1, 3).contiguous()
        attn_output = attn_output.view(batch_size, -1, self.d_model)

        output = self.out_linear(attn_output)
        output = output.permute(0, 2, 1)
        return output




import numpy as np
import torch.nn.functional as F
import os
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt


class UEA(Dataset):
  def __init__(self, name,device,augmentation = False,classes=None, **kwargs):
    if "path" in kwargs and kwargs["path"] != None:
      self.PATH = kwargs["path"]
    else:
      self.PATH = "/home/cmackinnon/scripts/datasets/UEA_NPY/"
    self.augmentation = augmentation
    ##LOAD SAMPLES AND LABELS FROM .npy FILE
    x = []
    y = []
    self.sizes = []

    for n in name:
      x.append(np.load("{}{}_samples.npy".format(self.PATH,n)))
      self.sizes.append(x[-1].shape[0])
      y.append(np.load("{}{}_labels.npy".format(self.PATH,n)))
      
      if "{}_groups.npy".format(n) in os.listdir(self.PATH):
        self.groups = np.load("{}{}_groups.npy".format(self.PATH,n))
    self.x = torch.from_numpy(np.concatenate(x,axis = 0)).to(device = device).float()
    self.x = torch.nan_to_num(self.x)
    self.y = np.concatenate(y,axis = 0)
    if kwargs["binary"]:
      self.y = np.where(self.y != 0, 1,0)
    self.y = torch.from_numpy(self.y).to(device).long()
    
    if classes != None:
        """
        for c in classes:
            np.where(self.y == c)
        """
    self.n_classes = len(torch.unique(self.y))
    
    self.n_features = self.x.shape[1]
  def __getitem__(self,index):
    x ,y = self.x[index], self.y[index]
    if self.augmentation:
      for f in self.augmentation:
        x,y = f(x,y)
    return x,y

  def update_device(self,device):
    self.x.to(device)
    self.y.to(device)
  def __len__(self):
    return len(self.y)
  def enable_augmentation(self,augs):
    self.augmentation = augs
  def min_samples_per_class(self):
    unique,counts = np.unique(self.y.cpu().numpy(), return_counts=True)
    return min(counts)
  def disable_augmentation(self):
    self.augmentation = False 
  def get_groups(self, train_id, test_id):
      train_groups = self.groups[train_id]
      test_groups = self.groups[test_id]
      #print("Groups in train are: {} with a total length: {}".format(np.unique(train_groups),len(train_groups)))
      #print("Groups in test are: {} with a total length: {}".format(np.unique(test_groups),len(test_groups)))
  def get_n_classes(self):
    return self.n_classes
  def get_n_features(self):
    return self.n_features
  def get_length(self):
    return self.x.shape[2]

  def get_proportions(self):
    return self.sizes[1]/ sum(self.sizes)

class UEA_Train(UEA):
  def __init__(self, name,device,**kwargs):
    name = "{}_{}".format(name,"train")
    super(UEA_Train,self).__init__(name = [name], device = device,**kwargs)
    
class UEA_Test(UEA):
  def __init__(self, name,device,**kwargs):
    name = "{}_{}".format(name,"test")
    super(UEA_Test,self).__init__(name = [name], device = device)

class UEA_Full(UEA):
  def __init__(self, name,device,**kwargs):
    name = ["{}_{}".format(name,"train") , "{}_{}".format(name,"test")]
    super(UEA_Full,self).__init__(name = name, device = device,**kwargs)

class Train_CharacterTrajectories(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("CharacterTrajectories","train")
    super(Train_CharacterTrajectories,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class Test_CharacterTrajectories(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("CharacterTrajectories","test")
    super(Test_CharacterTrajectories,self).__init__(name = [name], device = cuda_device,**kwargs)


class Train_LSST(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("LSST","train")
    super(Train_LSST,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class Test_LSST(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("LSST","test")
    super(Test_LSST,self).__init__(name = [name], device = cuda_device,**kwargs)

class Full_LSST(UEA):
  def __init__(self, cuda_device,**kwargs):
    name = ["{}_{}".format("LSST","train") , "{}_{}".format("LSST","test")]
    super(Full_LSST,self).__init__(name = name, device = cuda_device,**kwargs)

class Validation_LSST(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("LSST","validation")
    super(Validation_LSST,self).__init__(name = [name], device = cuda_device,**kwargs)
    


class Train_PhonemeSpectra(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("PhonemeSpectra","train")
    super(Train_PhonemeSpectra,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class Test_PhonemeSpectra(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("PhonemeSpectra","test")
    super(Test_PhonemeSpectra,self).__init__(name = [name], device = cuda_device,**kwargs)

class Full_PhonemeSpectra(UEA):
  def __init__(self, cuda_device,**kwargs):
    name = ["{}_{}".format("PhonemeSpectra","train") , "{}_{}".format("PhonemeSpectra","test")]
    super(Full_PhonemeSpectra,self).__init__(name = name, device = cuda_device,**kwargs)
 

class Train_EthanolConcentration(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("EthanolConcentration","train")
    super(Train_EthanolConcentration,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class Test_EthanolConcentration(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("EthanolConcentration","test")
    super(Test_EthanolConcentration,self).__init__(name = [name], device = cuda_device,**kwargs)
class Full_EthanolConcentration(UEA):
  def __init__(self, cuda_device,**kwargs):
    name = ["{}_{}".format("EthanolConcentration","train") , "{}_{}".format("EthanolConcentration","test")]
    super(Full_EthanolConcentration,self).__init__(name = name, device = cuda_device,**kwargs)


    

class Train_PenDigits(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("PenDigits","train")
    super(Train_PenDigits,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class Test_PenDigits(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("PenDigits","test")
    super(Test_PenDigits,self).__init__(name = [name], device = cuda_device,**kwargs)

class Validation_PenDigits(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("PenDigits","validation")
    super(Validation_PenDigits,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class True_Test_PenDigits(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("PenDigits","true_test")
    super(True_Test_PenDigits,self).__init__(name = [name], device = cuda_device,**kwargs)

class Full_PenDigits(UEA):
  def __init__(self, cuda_device,**kwargs):
    name = ["{}_{}".format("PenDigits","train") , "{}_{}".format("PenDigits","test")]
    super(Full_PenDigits,self).__init__(name = name, device = cuda_device,**kwargs)

class Validation_PhonemeSpectra(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("PhonemeSpectra","validation")
    super(Validation_PhonemeSpectra,self).__init__(name = [name], device = cuda_device,**kwargs)


class Train_FaceDetection(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("FaceDetection","train")
    super(Train_FaceDetection,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class Test_FaceDetection(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("FaceDetection","test")
    super(Test_FaceDetection,self).__init__(name = [name], device = cuda_device,**kwargs)

class Full_FaceDetection(UEA):
  def __init__(self, cuda_device,**kwargs):
    name = ["{}_{}".format("FaceDetection","train") , "{}_{}".format("FaceDetection","test")]
    super(Full_FaceDetection,self).__init__(name = name, device = cuda_device,**kwargs)

class Validation_FaceDetection(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("FaceDetection","val")
    super(Validation_FaceDetection,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class True_Test_FaceDetection(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("FaceDetection","true_test")
    super(True_Test_FaceDetection,self).__init__(name = [name], device = cuda_device,**kwargs)




class Train_UWaveGestureLibrary(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("UWaveGestureLibrary","train")
    super(Train_UWaveGestureLibrary,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class Test_UWaveGestureLibrary(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("UWaveGestureLibrary","test")
    super(Test_UWaveGestureLibrary,self).__init__(name = [name], device = cuda_device,**kwargs)

class Full_UWaveGestureLibrary(UEA):
  def __init__(self, cuda_device,**kwargs):
    name = ["{}_{}".format("UWaveGestureLibrary","train") , "{}_{}".format("UWaveGestureLibrary","test")]
    super(Full_UWaveGestureLibrary,self).__init__(name = name, device = cuda_device,**kwargs)
 

class Train_SpokenArabicDigits(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("SpokenArabicDigits","train")
    super(Train_SpokenArabicDigits,self).__init__(name = [name], device = cuda_device,**kwargs)
    
class Test_SpokenArabicDigits(UEA):
  def __init__(self,cuda_device,**kwargs):
    name = "{}_{}".format("SpokenArabicDigits","test")
    super(Test_SpokenArabicDigits,self).__init__(name = [name], device = cuda_device,**kwargs)

class Full_SpokenArabicDigits(UEA):
  def __init__(self, cuda_device,**kwargs):
    name = ["{}_{}".format("SpokenArabicDigits","train") , "{}_{}".format("SpokenArabicDigits","test")]
    super(Full_SpokenArabicDigits,self).__init__(name = name, device = cuda_device,**kwargs)



class Train_N(UEA):
  def __init__(self,ds,cuda_device,**kwargs):
    name = "{}_{}".format(ds,"train")
    super(Train_N,self).__init__(name = [name], device = cuda_device,**kwargs)

class Retrain_N(UEA):
  def __init__(self,ds,cuda_device,**kwargs):
    name = ["{}_{}".format(ds,"train"),"{}_{}".format(ds,"test")]
    super(Retrain_N,self).__init__(name = name, device = cuda_device,**kwargs)
    
class Test_N(UEA):
  def __init__(self,ds,cuda_device,**kwargs):
    name = "{}_{}".format(ds,"test")
    super(Test_N,self).__init__(name = [name], device = cuda_device,**kwargs)

class Full_N(UEA):
  def __init__(self,ds, cuda_device,**kwargs):
    name = ["{}_{}".format(ds,"train") , "{}_{}".format(ds,"test")]
    super(Full_N,self).__init__(name = name, device = cuda_device,**kwargs)

class Validation_N(UEA):
  def __init__(self,ds,cuda_device,**kwargs):
    name = "{}_{}".format(ds,"validation")
    super(Validation_N,self).__init__(name = [name], device = cuda_device,**kwargs)


import torch.nn as nn
import math
import torch
import numpy as np
from HPO.utils.graph_ops import OPS, ACTIVATION , NORMALISATION
import copy
import networkx as nx 
from HPO.utils.graph_utils import get_reduction, Order, get_sorted_edges
import matplotlib.pyplot as plt

def propagate_channels(edges,ops):
  down_sample_nodes = []
  g = nx.DiGraph()
  g.add_edges_from(edges)
  for i in ops:
    if "channel" in i:
        if ops[i] != 1:
          ##print(i,ops[i])
          down_sample_nodes.append((int(i.split("_")[0])))
  res_dict = {}
  for n in g.nodes():
      res_dict[n] = 1
  nodes = list(nx.topological_sort(g))
  for i in nodes:
      if i in down_sample_nodes:
          update_list = list(nx.bfs_tree(g, source=i).nodes())
          for _nodes in update_list:
              res_dict[_nodes] *= ops["{}_channel_ratio".format(i)]
  #plt.figure(figsize = (19,12))
  #nx.draw(
  #    g, edge_color='black', width=1, linewidths=1,
  #    node_size=500, node_color='pink', alpha=0.9,
  #    labels={node: "{}({})".format(node,res_dict[node]) for node in g.nodes()}
  #    )
  #plt.axis('off')
  #plt.savefig("channels")
  return res_dict

def propagate_channels_combine(edges,ops):
  down_sample_nodes = []
  combine_nodes = []
  g = nx.DiGraph()
  g.add_edges_from(edges)
  for i in ops:
    if "channel" in i and i.split("_")[0].isdigit():
          down_sample_nodes.append((int(i.split("_")[0])))
    if "combine" in i:
          if i.split("_")[0].isdigit(): 
            combine_nodes.append((int(i.split("_")[0])))
          else:
            combine_nodes.append((i.split("_")[0]))
  res_dict = {}
  for n in g.nodes():
      res_dict[n] = 1
  nodes = list(nx.topological_sort(g))
  for i in nodes:
    if i in combine_nodes:
      if ops["{}_combine".format(i)] == "CONCAT":
          predecessor_nodes = list(g.predecessors(i))
          ##print("Combine values: {}".format(predecessor_nodes))
          if len(predecessor_nodes) > 1:
            res_dict[i] *= len([res_dict[x] for x in predecessor_nodes])
      else:
        predecessor_nodes = list(g.predecessors(i))
        if len(predecessor_nodes ) == 0:
          res_dict[i] = 1 
        else:
          res_dict[i]  = max([res_dict[x] for x in predecessor_nodes])
    if i in down_sample_nodes:
      res_dict[i] *= ops["{}_channel_ratio".format(i)]
      #    update_list = list(nx.bfs_tree(g, source=i).nodes())
      #    update_list.remove(i)
      #    #print("Down Stream Nodes: {}".format(update_list))
      #    for _nodes in update_list:

      ##print(res_dict)
  ##print("combines: {}".format(res_dict))
  ##print(ops)
  return res_dict


def propagate_cat_num(edges,ops):
  down_sample_nodes = []
  combine_nodes = []
  g = nx.DiGraph()
  g.add_edges_from(edges)
  for i in ops:
    if "channel" in i and i.split("_")[0].isdigit():
          down_sample_nodes.append((int(i.split("_")[0])))
    if "combine" in i:
          if i.split("_")[0].isdigit(): 
            combine_nodes.append((int(i.split("_")[0])))
          else:
            combine_nodes.append((i.split("_")[0]))
  res_dict = {}
  for n in g.nodes():
      res_dict[n] = 1
      ##print(n)
  nodes = list(nx.topological_sort(g))
  for i in nodes:
    if i in combine_nodes:
      ##print("Node: {}".format(i))
      if ops["{}_combine".format(i)] == "CONCAT":
          predecessor_nodes = list(g.predecessors(i))
          #print("Combine values: {}".format(predecessor_nodes))
          if len(predecessor_nodes) > 1:
            res_dict[i] = len([res_dict[x] for x in predecessor_nodes])
    
  return res_dict


def propagate_resolution(edges, ops):
  ##print(edges)
  down_sample_nodes = []
  g = nx.DiGraph()
  g.add_edges_from(edges)
  for i in ops:
    if "stride" in i:
      if ops[i] > 1:
        ##print(i,ops[i])
        down_sample_nodes.append(int(i.split("_")[0]))
  res_dict = {}
  for n in g.nodes():
      res_dict[n] = 1
  nodes = list(nx.topological_sort(g))
  for i in nodes:
      if i in down_sample_nodes:
          update_list = list(nx.bfs_tree(g, source=i).nodes())
          for _nodes in update_list:
              res_dict[_nodes] *= ops["{}_stride".format(i)]
  """
  plt.figure(figsize = (19,12))
  nx.draw(
      g, edge_color='black', width=1, linewidths=1,
      node_size=500, node_color='pink', alpha=0.9,
      labels={node: "{}({})".format(node,res_dict[node]) for node in g.nodes()}
      )

  plt.axis('off')
  plt.savefig("resolutions")
  """
  return res_dict

class Node:
  """
  Resolution should never increase here
  """
  def __init__(self,name,data,length,channels,cat_num):
    self.name = name
    self.combine = data["{}_combine".format(name)]
    self.length = length
    self.cat_num = cat_num
    self.channels = int(channels)
    self.activation = data["{}_activation".format(name)]
    self.normalisation = data["{}_normalisation".format(name)]

  def generate_edge(self,node_previous) -> list: 
    ops = []#nn.ModuleList()
    ##print("output channel sizes : {} {} {}".format(self.channels, self.cat_num,self.channels/self.cat_num))
    if self.channels/self.cat_num != node_previous.channels:
      ops.append(OPS["resample_channels"](node_previous.channels,int(self.channels/self.cat_num)))
    self.stride = 2**round(math.log(node_previous.length // self.length,2))
    #if self.stride > 513:
      #print("lengths:",node_previous.length,self.length,self.stride)
    if self.stride != 1:
      ops.append(OPS["downsample_resolution"](int(self.channels/self.cat_num),self.stride))
    if self.normalisation != "none":
      ops.append(NORMALISATION[self.normalisation](int(self.channels/self.cat_num),self.length)) 
    if self.activation != "none":
      ops.append(ACTIVATION[self.activation](int(self.channels/self.cat_num),self.length)) 
    return ops
    


def transform_idx(original_list,original_list_permuted,new_list):
  """
  Takes in 3 arrays of the same shape and elements, an list_array which has been sorted in some way
  to make a new original_list_permuted and applies the same transform to the new_list as has been done 
  to the original_array.
  """
  transform_idx = [original_list_permuted.index(i) for i in original_list]
  return list(np.asarray(new_list)[transform_idx])

class ModelGraph(nn.Module):
  def __init__(self,n_features, n_channels, n_classes,signal_length, 
    graph : list, ops : list, device,binary = False,data_dim = 1,sigmoid = False,
    dropout = 0.3,droppath = False,raw_stem = False,embedding = False):
    super(ModelGraph,self).__init__()
    #INITIALISING MODEL VARIABLES
    self.DEBUG = False
    self.device = device
    self.data_dim = data_dim
    self.n_features = n_features
    self.graph = copy.copy(graph)
    self.sorted_graph = get_sorted_edges(graph)
    self.current_iteration = -1
    self.n_channels = n_channels
    self.ops_list = ops

    #STRUCTURES FOR HOLDING DATA STATES AND OPERATIONS
    self.states = {}
    self.nodes  = {}
    if dropout and droppath:
      self.dropout = DropPath(dropout,self.device)
    elif dropout:
      self.dropout = nn.Dropout(dropout)
    else:
      self.dropout = nn.Identity()
    self.ops = nn.ModuleDict()
    self.combine_ops = nn.ModuleDict() 
    self.embedding = embedding
    #BUILD STEM
    """
    This is a way of increasing the capacity of a model but upping the resolution of the image off the bat.
    Not sure why the stride is 2 but that was just how it was implemented in the efficient-net implementation I saw
    Probably more useful for image data honestly, so will probably just move it inside the if statement for now
    """
    if data_dim == 2:
      STEM_PADDING = 32
      STEM_STRIDE = 2 
      self.stem = nn.Conv2d(n_features,n_channels,2,stride = STEM_STRIDE ,padding = STEM_PADDING)
    else:
      if self.embedding:
        self.stem = nn.Embedding(258,n_channels)
      else:
        if raw_stem == True:
          self.stem = nn.Conv1d(n_features,n_channels,1,stride = 1)
        else:
          self.stem = nn.Conv1d(n_features,n_channels,2,stride = 2) #Will just leave this at defaults for now
    self.stem = self.stem.cuda(device)

    #DEFINE OP_MODULE BASED ON DATA_DIM
    #if self.data_dim == 2:
    #  self.OP_MODULE = OPS2D
    #else:
    #  self.OP_MODULE = OPS1D
  
    #BUILD NODES
    self.resolution_dict = propagate_resolution(self.sorted_graph, self.ops_list)
    self.channel_dict = propagate_channels(self.sorted_graph, self.ops_list)
    self.cat_num_dict = propagate_cat_num(self.sorted_graph, self.ops_list)
    ##print(self.channel_dict)
    ##print(self.resolution_dict)
    if not raw_stem:
      signal_length = signal_length//2
    g = nx.DiGraph()
    g.add_edges_from(self.sorted_graph)
    for i in g.nodes():
      self.nodes[i] = Node(i,self.ops_list,signal_length // self.resolution_dict[i], self.channel_dict[i]*n_channels,self.cat_num_dict[i])

    #BUILDS THE OPERATIONS ALONG EDGES BASED ON N_CHANNELS OF PREVIOUS OP
    self._compile(signal_length)
       

    #BUILD CLASSIFIER
    C = self.nodes["T"].channels
    if data_dim == 2:
      self.global_pooling = nn.AdaptiveAvgPool2d(1)
    else:
      self.global_pooling = nn.AdaptiveAvgPool1d(1)
    if binary == True:
      self.binary = binary
      self.classifier = nn.Linear(C, 1)
    else:
      self.classifier = nn.Linear(C, n_classes)
    if sigmoid:
      self.actfc = nn.Sigmoid()#nn.Softmax(dim =1)
    self.sigmoid = sigmoid
  
  def _compile(self,size):
    """
    Builds the operations along edge paths, 
    """
    #GENERATE RANDOM DATA TO PASS THROUGH THE NETWORK
    batch = 16
    self.combine_index = 0
    if self.embedding:
      x = torch.zeros(size = (batch, self.n_features,size)).long().cuda(self.device)
    else:
      if self.data_dim == 2:
        x = torch.rand(size = (batch, self.n_features,size,size)).cuda(self.device)
      else:
        x = torch.rand(size = (batch, self.n_features,size)).cuda(self.device)      
    x = self.stem(x)

    self.states["S"] = x
    
    #ITERATE THROUGH EDGES
    for iteration,edge in enumerate(self.sorted_graph):
        #print(edge, self.ops_list["{}_{}_OP".format(edge[0],edge[1])])
        #print(self.channel_dict[edge[0]]*self.n_channels,self.channel_dict[edge[1]]*self.n_channels)
      
      #INITIALISE CONTAINER FOR EDGE OPERATIONS 
      edge_container = nn.Sequential()

      #GET OP AND NODE DATA
      curr_op = self.ops_list["{}_{}_OP".format(edge[0],edge[1])]
      curr_node = self.nodes[edge[1]]
      node_ops = curr_node.generate_edge(self.nodes[edge[0]])

      #GET NUMBER OF CHANNELS FROM PREVIOUS DATA STATE
      if edge[0] == "S":#INIT CHANNELS
        C = self.n_channels
      else:
        C = self.nodes[edge[0]].channels
      
      #BUILD THE OPERATION
      if curr_op[:2] == "L_":
        edge_container.append( OPS[curr_op](C,self.nodes[edge[0]].length).cuda(self.device))
      else:
        edge_container.append( OPS[curr_op](C).cuda(self.device))
      for i in node_ops:
        edge_container.append(i) 
       

      #ADD OP TO THE MODULE LIST AND PASS THROUGH DATA
      self.ops[str(edge)] = edge_container
      

  def combine(self,x1,x2):
    """
    Accepts two tensors [B,C,L] and returns [B,C,L1,L2]
    """

    batch_size  = x1.shape[0]
    channels = x1.shape[1]
    out = torch.bmm(x1.view(-1,x1.shape[-1],1),x2.view(-1,1,x2.shape[-1]))
    proper_size_out = out.view(batch_size,channels,out.shape[-2],out.shape[-1])
    return proper_size_out

  def next_op(self):
    self.current_iteration +=1
    op, edge = self.ops[self.current_iteration] , self.edges[self.current_iteration]
    return self.current_iteration, op, edge 
  
  def _forward_build(self,op,edge,iteration):
    h = op(self.states[edge[0]])
    #CASE 1 - 1 INPUT
    if not (edge[1] in self.states.keys()):
      self.states[edge[1]] = h
    #CASE 2 - 2 INPUTS OF SAME SIZE (ADD)
    elif self.states[edge[1]].shape == h.shape:
      self.states[edge[1]] = self.states[edge[1]] + h
    #CASE 3 - 2 INPUTS, SAME LENGTH (CONCAT CHANNELS)
    elif self.states[edge[1]].shape[2] == h.shape[2] and False:
      self.states[edge[1]] = torch.cat((self.states[edge[1]], h),dim = 1)
    #CASE 4 - 2 INPUTS SAME CHANNELS (MATMUL 2D CONV)
    elif self.states[edge[1]].shape[1] == h.shape[1] and False:
      h = self.combine(self.states[edge[1]], h)
      if not self.combine_index in self.combine_ops:
        #BUILD KERNEL OF SIZE L1,L2 THEN SETS LARGER DIM TO 3
        kernel = torch.tensor(h.shape[-2:])
        channels = h.shape[1]
        kernel[torch.argmax(kernel)] = 3
        self.combine_ops[str(edge)] = nn.Conv2d(channels,channels,kernel,groups = channels).cuda(self.device)
      self.states[edge[1]] = self.combine_ops[str(edge)](h).squeeze()
      self.combine_index+=1
      
    #CASE 5 - DIFFERENT C AND L (SE OPERATION)
    else:
      channels_in = h.shape[1]
      channels_out = self.states[edge[1]].shape[1]
      self.combine_ops[str(edge)] = (self.OP_MODULE.SEMIX(channels_in,channels_out)).cuda(self.device)
      self.states[edge[1]] = self.combine_ops[str(edge)](self.states[edge[1]],h)
      self.combine_index+=1

  def _forward(self, op,edge):
    self.states[edge[0]] = self.dropout(self.states[edge[0]])
    ##print(edge,self.states[edge[0]].shape, op)
    ##print("Predicted channels: {} {}".format(self.channel_dict[edge[0]],self.channel_dict[edge[1]]))
    h = op(self.states[edge[0]])
    #CASE 1 - 1 INPUT
    if not (edge[1] in self.states.keys()):
      self.states[edge[1]] = h
    #CASE 2 - 2 INPUTS OF SAME SIZE (ADD)
    elif self.nodes[edge[1]].combine == "ADD":
      """
      if self.states[edge[1]].shape != h.shape:
        #print("PREADD SHAPE: {} - {}".format(self.states[edge[1]].shape, h.shape))
        #print("OP: {}".format(op))
        #print("PREVIOUS SHAPE: {}".format(self.states[edge[0]].shape))
      """
      self.states[edge[1]] = self.states[edge[1]] + h
    elif self.nodes[edge[1]].combine == "CONCAT":
      ##print("PRECONCAT: {} - {}".format(self.states[edge[1]].shape, h.shape))
      self.states[edge[1]] = torch.cat((self.states[edge[1]], h),dim = 1)
    elif self.nodes[edge[1]].combine == "MULT":
      self.states[edge[1]] = self.states[edge[1]] * h
    ##print(edge,self.states[edge[1]].shape)
    else:
      self.states[edge[1]] = self.states[edge[1]] + h
  def forward(self,x):
    self.combine_index = 0
    self.states = {}
    if self.embedding:
      x = x.squeeze().long()
      self.states["S"] = self.stem(x).permute(0,2,1)
    else: 
      self.states["S"] = self.stem(x)
    self.current_iteration = -1
    hold = 0
    #while self.current_iteration < len(self.edges)-1:
    for iteration, edge in enumerate(self.sorted_graph):
      op = self.ops[str(edge)]
      self._forward(op,edge)
    #FC LAYER
    x = self.global_pooling(self.states["T"])
    x = self.classifier(x.squeeze())
    if self.sigmoid:
      x = self.actfc(x)
    return x

class DropPath(nn.Module):
  def __init__(self, drop_prob,device):
    super(DropPath, self).__init__()
    self.drop_prob = drop_prob
    self.device = device
  def forward(self,x):
    if self.drop_prob > 0.:
      keep_prob = 1.-self.drop_prob
      mask = torch.FloatTensor(x.size(0), 1, 1).bernoulli_(keep_prob).cuda(self.device)
      #x.div(keep_prob)
      x.mul(mask)
    return x


      
    
    
